from hippodamia.hierarchivalview.treeentry import Entry
from hippodamia.hierarchivalview.treeleaf import Leaf
from hippodamia.enums import ViewDetails
from hippodamia.enums import Health
from collections import deque


class Branch(Entry):
    children = None

    def __init__(self, name):
        Entry.__init__(self, name)
        self.children = {}
        self.update()

    def update(self, children=None):
        if children is None:
            children = self.children

        self.health = Health.GREEN
        for gid, leaf in children.items():
            leaf_health = leaf.update()
            self.health = max(self.health, leaf_health)

        return self.health

    def add_leaf(self, leaf, breadcrumbs, children=None):
        if children is None:
            children = self.children

        if len(breadcrumbs) == 0:
            children[leaf.gid] = leaf
        else:
            branch_name = breadcrumbs.popleft()
            if branch_name is None or branch_name == "":
                branch_name = "_"

            if branch_name not in children:
                children[branch_name] = Branch(branch_name)
            children[branch_name].add_leaf(leaf, breadcrumbs)

    def pformat(self, details=ViewDetails.NONE, filters=None):
        if filters is None:
            filters = deque()
        lines = self.tree_view("", "", details, filters)
        result = "\n".join(lines)
        return  result

    @staticmethod
    def _filter_match(name, entry_filter):
        if entry_filter is not None:
            name = name.lower()
            entry_filter = entry_filter.lower()
            return name.startswith(entry_filter)
        return True

    def tree_view(self, prefix_first_line, prefix_next_lines, details, filters):
        filters = filters.copy()
        try:
            entry_filter = filters.popleft()
        except IndexError:
            entry_filter = None

        first_line = prefix_first_line + self._pprint_details(details)
        lines = [first_line]
        prefix_a = prefix_next_lines + self.TYPE_A
        prefix_b = prefix_next_lines + self.TYPE_B
        prefix_c = prefix_next_lines + self.TYPE_C
        prefix_d = prefix_next_lines + self.TYPE_D

        counter = 0
        for name, entry in self.children.items():
            counter += 1
            if self._filter_match(name, entry_filter):
                if counter == len(self.children):
                    # last entry
                    prefix_first = prefix_c
                    prefix_next = prefix_d
                else:
                    prefix_first = prefix_a
                    prefix_next = prefix_b

                if type(entry) is Leaf:
                    line = prefix_first + entry.pformat(details)
                    lines.append(line)
                elif type(entry) is Branch:
                    sub_lines = entry.tree_view(prefix_first, prefix_next, details, filters)
                    lines += sub_lines
                else:
                    raise TypeError

        return lines

    def _pprint_details(self, details):
        result = [self._name2str()]
        if details & ViewDetails.HEALTH == ViewDetails.HEALTH:
            result.append(self.health.name)
        if details & ViewDetails.STATE == ViewDetails.STATE:
            pass
        if details & ViewDetails.GID == ViewDetails.GID:
            pass
        result = " ".join(result)
        return result

    def to_dict(self):
        result = {
            "type": self.__class__.__name__,
            "name": self._name2str(),
            "health": self.health.name,
            "state": None,
            "gid": None,
            "children": []
        }
        for child in self.children.values():
            result["children"].append(child.to_dict())
        return result



from pelops.logging.mylogger import get_child
from asyncscheduler import AsyncScheduler


class ANotifier:
    _counter = None

    _config = None
    _logger = None
    _send_email = None

    _to = None
    _subject = None
    _pre_text = None
    _post_text = None

    _sched = None

    def __init__(self, config, send_email, logger):
        self._config = config
        self._logger = get_child(logger, self.__class__.__name__)
        self._logger.info("__init__")
        self._logger.debug("__init__ - config: {}".format(self._config))

        self._to = self._config["to"]
        self._subject = self._config["subject"]
        try:
            self._pre_text = self._config["pre-text"]
        except KeyError:
            self._pre_text = ""
        try:
            self._post_text = self._config["post-text"]
        except KeyError:
            self._post_text = ""

        self._send_email = send_email
        self._sched = AsyncScheduler()
        self._counter = 0

    def _send_message(self, body):
        self._logger.info("_send_message")
        message = "<html><body><div>{}</div><div>{}</div><div>{}</div><div>{}</div></body></html>"\
            .format(self._pre_text, body, self._post_text, self._message_footer())
        self._logger.debug("_send_message - message: {}".format(message.encode("utf-8")))
        self._send_email(self._subject, message, self._to)
        self._counter += 1

    def _message_footer(self):
        raise NotImplementedError()

    def start(self):
        self._logger.info("start")
        self._sched.start()
        self._start()

    def _start(self):
        raise NotImplementedError()

    def stop(self):
        self._logger.info("stop")
        self._sched.stop(wait=False)
        self._stop()

    def _stop(self):
        raise NotImplementedError()

    def get_stats(self):
        stats = {
            "count": self._counter,
            "to": self._to,
            "scheduled events": len(self._sched.scheduler.queue)
        }
        stats.update(self._get_stats())
        return stats

    def _get_stats(self):
        raise NotImplementedError()




from hippodamia.enums import Health, Necessity
import logging


class Properties:
    _logger = None

    gid = None
    necessity = None
    type = None
    module = None
    name = None
    location = None
    room = None
    device = None
    uuid = None
    onboarding_topic = None
    session = None
    protocol_version = None
    version = None
    description = None
    host_name = None
    node_id = None
    ips = None
    config_hash = None
    service_uptime = None
    state_health = None
    log_health = None
    time = None

    def __init__(self, logger):
        self._logger = logger

    def get_stats(self):
        try:
            health = self.health.name
        except AttributeError:
            health = None
        try:
            necessity = self.necessity.name
        except AttributeError:
            necessity = None

        result = {
            "time": self.time,
            "entries": {
                "gid": self.gid,
                "session": self.session,
                "health": {
                    "overall": health,
                    "state": self.state_health.name,
                    "log": self.log_health.name
                },
                "necessity": necessity,
                "uuid": self.uuid,
                "onboarding_topic": self.onboarding_topic,
                "protocol_version": self.protocol_version,
                "type": self.type,
                "module": self.module,
                "version": self.version,
                "name": self.name,
                "location": self.location,
                "room": self.room,
                "device": self.device,
                "description": self.description,
                "host_name": self.host_name,
                "node_id": self.node_id,
                "ips": self.ips,
                "config_hash": self.config_hash,
                "service_uptime": self.service_uptime
            }
        }
        return result

    def _get_health(self):
        try:
            result = max(self.log_health, self.state_health)
        except TypeError:
            try:
                result = max(self.log_health)
            except TypeError:
                try:
                    result = max(self.state_health)
                except TypeError:
                    result = Health.RED
        return result

    health = property(fget=_get_health)

    def set_state_health(self, health):
        if self.necessity == Necessity.REQUIRED:
            self.state_health = health
        else:
            self.state_health = min(Health.YELLOW, health)
        self._logger.debug("set_state_health to {} (original: {})"
                           .format(self.state_health.name, health.name))

    def set_log_health(self, max_level):
        if max_level >= logging.ERROR:
            self.log_health = Health.RED
        else:
            self.log_health = Health.GREEN
        self._logger.debug("set_log_health to {} (level: {})"
                           .format(self.log_health.name, logging.getLevelName(max_level)))



from tantamount.machine import Machine
import collections
import datetime
import copy


class MachineLogger(Machine):
    logger = None
    history = None
    _observer = None
    _gid = None

    def __init__(self, gid, logger):
        self.logger = logger.getChild(self.__class__.__name__)
        self.history = collections.deque(maxlen=100)
        Machine.__init__(self, ignore_undefined_events=True)
        self.registertransitionhook(self._hook)
        self._gid = gid

    def register_state_transistion_event_observer(self, observer):
        self._observer = observer

    def _hook(self, start_id, transition_id, target_id):
        timestamp = datetime.datetime.now()
        self.history.append({"from": start_id.name, "event": transition_id.name, "to": target_id.name, "time": timestamp})
        if self._observer:
            self._observer(self._gid, start_id, target_id)

    def operate(self, eventid, source_stateid=None, source_activation_counter=None):
        self.logger.debug("operate - eventid: {}".format(eventid))
        try:
            Machine.operate(self, eventid, source_stateid, source_activation_counter)
        except KeyError as err:
            self.logger.info("operate - KeyError: {}".format(err))
        except Exception as e:
            self.logger.error("operate - Exception: {}".format(e))
            raise e

    def asyncoperate(self, eventid, source_stateid=None, source_activation_counter=None):
        self.logger.debug("asyncoperate - eventid: {}".format(eventid))
        Machine.asyncoperate(self, eventid, source_stateid, source_activation_counter)

    def addstate(self, state):
        self.logger.debug("addstate - state: {}, stateid: {}, groupid: {}".format(state, state.id, state.groupid))
        Machine.addstate(self, state)

    def addtimeoutevent(self, stateid, eventid, seconds):
        self.logger.debug("addtimeoutevent - stateid: {}, eventid: {}, seconds: {}".format(stateid, eventid, seconds))
        Machine.addtimeoutevent(self, stateid, eventid, seconds)

    def addtransition(self, startstateid, transitionid, targetstateid, actionFunc=None, actionArgs=None):
        self.logger.debug("addtransition - startstateid: {}, transitionid: {}, targetstateid: {}, actionFunc=None: {}, "
                          "actionArgs: {}".format(startstateid, transitionid, targetstateid, actionFunc, actionArgs))
        Machine.addtransition(self, startstateid, transitionid, targetstateid, actionFunc, actionArgs)

    def start(self):
        self.logger.debug("start - start")
        Machine.start(self)
        self.logger.debug("start - finished")

    def asyncstart(self):
        self.logger.debug("asyncstart - start")
        Machine.asyncstart(self)
        self.logger.debug("asyncstart - finished")

    def get_active_state(self):
        self.logger.debug("get_active_state")
        return Machine.get_active_state(self)

    def restarttimeoutevent(self):
        self.logger.debug("restarttimeoutevent")
        Machine.restarttimeoutevent(self)

    def setstartstate(self, stateid):
        self.logger.debug("setstartstate - stateid: {}".format(stateid))
        Machine.setstartstate(self, stateid)

    def stop(self):
        self.logger.debug("stop - start")
        Machine.stop(self)
        self.logger.debug("stop - finished")

    def updatetimeoutevent(self, stateid, eventid, seconds):
        self.logger.debug("updatetimeoutevent - statid: {}, eventid: {}, seconds: {}".format(stateid, eventid, seconds))
        Machine.updatetimeoutevent(self, stateid, eventid, seconds)

    def _async_worker(self):
        self.logger.debug("_async_worker - start")
        try:
            Machine._async_worker(self)
        except RuntimeError as e:
            self.logger.error("_async_worker exited with: {}".format(e))
            raise e
        self.logger.debug("_asnyc_worker - finished")

    def get_stats_active_state(self):
        since = self._state.timestamp_on_entry
        try:
            duration = datetime.datetime.now() - since
        except TypeError:
            duration = None
        stats = {
            "output": self._state.id.name,
            "since": since,
            "duration": duration
        }
        return stats

    def get_stats(self):
        stats = {
            'active-state': self.get_stats_active_state(),
            'history': []
        }

        for h in self.history:
            entry = copy.deepcopy(h)
            entry["time"] = entry["time"].strftime("%Y-%m-%dT%H:%M:%S")
            stats["history"].append(entry)

        return stats



def get_schema():
    schema = {
        "type": "object",
        "description": "web page to view the state of the services",
        "properties": {
            "port": {
                "type": "integer",
                "example": "6060",
                "description": "web server port"
            },
            "webserver-user": {
                "description": "User name for webserver",
                "type": "string"
            },
            "webserver-password": {
                "description": "Password for webserver",
                "type": "string"
            },
            "credentials-file": {
                "type": "string",
                "example": "~/credentials.yaml",
                "description": "if credentials are set, website is password protected"
            },
            "log-level": {
                "type": "string",
                "example": "DEBUG",
                "description": "log level for mongodb client related messages",
                "enum": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
            }
        },
        "required": ["port", "log-level"],
        "additionalProperties": False
    }
    return schema


import pprint
from hippodamia.agentshadow.agentshadowfactory import AgentShadowFactory
import pelops.ui.tools
from hippodamia.enums import StatType
from hippodamia.enums import ViewDetails


class CommandFactory:
    _logger = None
    _shadows = None
    _hierarchical_view = None
    _request_ping = None
    _request_runtime = None
    _request_config = None
    _request_reonboarding = None
    _request_end = None
    _heartbeat = None
    _onboarding = None
    _config = None
    _archive = None
    _notification = None

    def __init__(self, version, config, shadows, heartbeat, health_view, onboarding, request_ping, request_runtime,
                 request_config, request_reonboarding, request_end, archive, notification, logger):
        self._shadows = shadows
        self._config = config
        self._hierarchical_view = health_view
        self._request_ping = request_ping
        self._request_runtime = request_runtime
        self._request_config = request_config
        self._request_reonboarding = request_reonboarding
        self._request_end = request_end
        self._logger = logger
        self._heartbeat = heartbeat
        self._onboarding = onboarding
        self._archive = archive
        self._notification = notification

    def add_commands(self, add_command):
        add_command("request_ping", self.do_request_ping)
        add_command("request_runtime", self.do_request_runtime)
        add_command("request_config", self.do_request_config)
        add_command("request_reonboarding", self.do_request_reonboarding)
        add_command("request_end", self.do_request_end)
        add_command("list", self.do_list)
        add_command("hierarchical", self.do_hierarchical)
        add_command("agentstats", self.do_agentstats)
        add_command("agentlog", self.do_agentlog)
        add_command("agentmessages", self.do_agentmessages)
        add_command("heartbeatstats", self.do_heartbeatstats)
        add_command("onboardingstats", self.do_onboardingstats)
        add_command("notificationstats", self.do_notificationstats)
        add_command("archive", self.do_archive)
        add_command("ad", self.do_ad)  # agentdetails full
        add_command("am", self.do_am)  # agentmessages
        add_command("al", self.do_al)  # agentmessages
        add_command("ag", self.do_ag)  # agentdetails general
        add_command("l", self.do_l)  # list
        add_command("h", self.do_h)  # hierarchical
        add_command("o", self.do_o)  # onboarding stats

    def _check_gid(self, gid):
        if gid not in self._shadows:
            raise KeyError("gid {} not in {}".format(gid, list(self._shadows.keys())))
        return gid

    def _request_extract_gid(self, name, arg):
        args = pelops.ui.tools.parse(arg)

        gid = None
        if len(args) == 1 and str(args[0]).lower() == "all":
            print("Sending {} request to all monitored services\n".format(name))
        elif len(args) == 2 and str(args[0]).lower() == "gid":
            gid = self._check_gid(args[1])
            print("Send {} regest to monitoring service with gid == {}\n".format(name, gid))
        else:
            raise ValueError("Wrong arguments: {}. expected '{} [all|gid gid]'.\n".format(args, name))

        return gid

    def do_exit(self, arg):
        """exit - stops the monitoring service: EXIT"""
        return True

    def do_request_ping(self, arg):
        """request_ping [all|gid gid] - publish ping request either to all monitored services or to a specific one: REQUEST_PING GID 1"""
        try:
            self._request_ping(self._request_extract_gid("ping", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))

    def do_request_runtime(self, arg):
        """request_runtime [all|gid gid] - publish runtime request either to all monitored services or to a specific one: REQUEST_RUNTIME GID 1"""
        try:
            self._request_runtime(self._request_extract_gid("runtime", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_config(self, arg):
        """request_config [all|gid gid] - publish config request either to all monitored services or to a specific one: REQUEST_CONFIG GID 1"""
        try:
            self._request_config(self._request_extract_gid("config", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_reonboarding(self, arg):
        """request_reonboarding [all|gid gid] - publish reonboarding request either to all monitored services or to a specific one: REQUEST_REONBOARDING GID 1"""
        try:
            self._request_reonboarding(self._request_extract_gid("reonboarding", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_end(self, arg):
        """request_end [all|gid gid] - publish end request either to all monitored services or to a specific one: REQUEST_END GID 1"""
        try:
            self._request_end(self._request_extract_gid("end", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))

    def do_list(self, arg):
        """list - list all known services: LIST"""
        m = AgentShadowFactory.get_string_list(self._shadows)
        pelops.ui.tools.more(m)

    def do_archive(self, arg):
        """archive - list all entries in archive: ARCHIVE"""
        m = "archived agentshadows:\n"
        for a in self._archive:
            m += " - [{}] {}\n".format(a.properties.gid, a.properties.name)
        pelops.ui.tools.more(m)

    def do_hierarchical(self, arg):
        """hierarchical [location location|room room|device device|service service]*
 - show all monitored services: HIERARCHICAL
 - show one location: HIERARCHICAL LOCATION Flat
 - show one specific service: HIERARCHICAL LOCATION Flat ROOM \"Living Room\" DEVICE Thermostat SERVICE \"Thermostat GUI\"
 - show all service that run on devices with the name 'Thermostat': HIERARCHICAL DEVICE Thermostat"""
        syntax = "hierarchical [location location|room room|device device|service service]*"
        arguments = pelops.ui.tools.parse(arg)
        if len(arguments) % 2 == 1:
            print("wrong number of arguments ({}). expected '{}'".format(len(arguments), syntax))
            return
        elif len(arguments) == 0:
            m = self._hierarchical_view.pformat(details=ViewDetails.FULL)
        else:
            filters = {"LOCATION": "", "ROOM": "", "DEVICE": "", "SERVICE": ""}
            pos = 0
            while pos < len(arguments):
                first = arguments[pos].upper()
                second = arguments[pos+1]
                if first in filters:
                    filters[first] = second
                else:
                    print("don't know a parameter/filter named '{}'. expected '{}'".format(first, syntax))
                    return
                pos += 2
            m = self._hierarchical_view.pformat(details=ViewDetails.FULL, filters=filters)
        pelops.ui.tools.more(m)

    def do_agentstats(self, arg):
        """agentstats gid gid [all|general|state_machine|properties|logger|messages]
        show information on one specific monitored service: AGENTSTATS GID 1 PROPERTIES"""
        syntax = "agentstats gid gid [all|general|state_machine|properties|logger|messages]"

        args = pelops.ui.tools.parse(arg)
        if len(args) != 3:
            print("wrong number of arguments: {}. expected '{}'\n".format(args, syntax))
        elif str(args[0]).lower() != "gid":
            print("first argument ({}) must be 'gid'. expected '{}'\n".format(str(args[0]).lower(), syntax))
        elif str(args[2]).upper() not in StatType.get_members():
            print("third argument ({}) must be one of {}. expected '{}'\n".format(str(args[2]).lower(),
                                                                                  StatType.get_members(), syntax))
        else:
            try:
                stat_type = StatType.get_enum(str(args[2]))
                gid = self._check_gid(args[1])
                output = "stats {} for gid {}:\n".format(type, gid)
                output += self._shadows[gid].get_string_stats(stat_type)
                pelops.ui.tools.more(output)
            except (ValueError, KeyError) as e:
                print("{}\n".format(e))

    def do_agentlog(self, arg):
        """agentlog gid gid [FLAT|DEBUG|INFO|WARNING|ERROR|CRITICAL] - view all received log-entries of this
service: AGENTLOG GID 1 FLAT"""
        syntax = "agentlog gid gid [FLAT|DEBUG|INFO|WARNING|ERROR|CRITICAL]"

        args = pelops.ui.tools.parse(arg)
        if len(args) != 3:
            print("wrong number of arguments: {}. expected '{}'\n".format(args, syntax))
        elif str(args[0]).lower() != "gid":
            print("first argument ({}) must be 'gid'. expected '{}'\n".format(str(args[0]).lower(), syntax))
        elif str(args[2].lower()) not in ["flat", "debug", "info", "warning", "error", "critical"]:
            print("unexpected value for third argument '{}'. expected '{}'\n".format(args[2], syntax))
        else:
            try:
                gid = self._check_gid(args[1])
                output = "show {} log for gid: {}\n".format(args[2], gid)
                output += self._shadows[gid].log_archive.get_string_stats(args[2])
                pelops.ui.tools.more(output)
            except KeyError as e:
                print("{}\n".format(e))

    def do_agentmessages(self, arg):
        """agentmessages gid gid [ALL|CONFIG|END|LOGGER|PING|REQUEST|RUNTIME] - view all received messages of this
service: AGENTMESSAGES GID 1 PING"""
        syntax = "agentmessages gid gid [ALL|CONFIG|END|LOGGER|PING|REQUEST|RUNTIME]"

        args = pelops.ui.tools.parse(arg)
        if len(args) != 3:
            print("wrong number of arguments: {}. expected '{}'\n".format(args, syntax))
        elif str(args[0]).lower() != "gid":
            print("first argument ({}) must be 'gid'. expected '{}'\n".format(str(args[0]).lower(), syntax))
        elif str(args[2].lower()) not in ["all", "config", "end", "logger", "ping", "request", "runtime"]:
            print("unexpected value for third argument '{}'. expected '{}'\n".format(args[2], syntax))
        else:
            try:
                gid = self._check_gid(args[1])
                key = args[2].lower()
                output = "show {} messages for gid: {}\n".format(key, gid)
                output += self._shadows[gid].message_archive.get_string_stats(key)
                pelops.ui.tools.more(output)
            except KeyError as e:
                print("{}\n".format(e))

    def do_heartbeatstats(self, arg):
        """heartbeatstats - stats on the heartbeat signal of the monitoring agent: HEARTBEATSTATS"""
        output = "heartbeat: \n"
        output += self._heartbeat.get_string_stats()
        pelops.ui.tools.more(output)

    def do_onboardingstats(self, arg):
        """onboardingstats - stats on the overall onboarding process: ONBOARDINGSTATS"""
        output = "onboarding: \n"
        output += self._onboarding.get_string_stats()
        pelops.ui.tools.more(output)

    def do_notificationstats(self, arg):
        """notificationstats - stats on the email notifcation subsystem: NOTIFICATIONSTATS"""
        output = "email notification: \n"
        output += self._notification.get_string_stats()
        pelops.ui.tools.more(output)

    def do_config(self, arg):
        """config - show the current config: CONFIG"""
        output = pprint.pformat(self._config, indent=2)
        pelops.ui.tools.more(output)

    def _get_default_gid(self):
        gid = "Monitoringservice-0"
        try:
            gid = self._check_gid(gid)
            return gid
        except KeyError:
            raise KeyError("unknown gid '{}'".format(gid))

    def do_ad(self, arg):
        # """shortcut for 'agentdetails gid Monitoringservice-0 all'"""
        args = "gid Monitoringservice-0 all"
        self.do_agentstats(args)

    def do_ag(self, arg):
        # """shortcut for 'agentdetails gid Monitoringservice-0 general'"""
        args = "gid Monitoringservice-0 general"
        self.do_agentstats(args)

    def do_am(self, arg):
        # """shortcut for 'agentmessages gid Monitoringservice-0 all'"""
        args = "gid Monitoringservice-0 all"
        self.do_agentmessages(args)

    def do_al(self, arg):
        # """shortcut for 'stats gid Monitoringservice-0 all'"""
        args = "gid Monitoringservice-0 flat"
        self.do_agentlog(args)

    def do_h(self, arg):
        # """shortcut for 'hierarchical'"""
        self.do_hierarchical(arg)

    def do_l(self, arg):
        # """shortcut for 'list'"""
        self.do_list(arg)

    def do_o(self, arg):
        # """ shortcut for 'onboardingstats'"""
        self.do_onboardingstats(arg)



from collections import deque
from hippodamia.hierarchivalview.treeroot import Root
from hippodamia.hierarchivalview.treeleaf import Leaf
from hippodamia.enums import ViewDetails
from threading import Lock
from pelops.logging.mylogger import get_child


class HierarchicalView:
    _logger = None
    _agentshadows = None

    tree = None
    _tree_lock = None

    def __init__(self, agentshadows, logger):
        self._logger = get_child(logger, self.__class__.__name__)
        self._logger.info("__init__")
        self._agentshadows = agentshadows
        self._tree_lock = Lock()
        self._update_tree()
        self._logger.debug("__init__ - finished")

    def _update_tree(self):
        self._logger.debug("_update_tree")
        self._build_tree()
        self._logger.debug("_update_tree - updating info within tree")
        self.tree.update()

    def pformat(self, details=ViewDetails.HEALTH, filters=None):
        self._logger.debug("pformat - waiting for lock")
        with self._tree_lock:
            self._update_tree()
            if filters is None:
                filters = deque([None, None, None, None])
            else:
                filters = deque([filters["LOCATION"], filters["ROOM"], filters["DEVICE"], filters["SERVICE"]])
            result = self.tree.pformat(details, filters)
            self._logger.info("pformat success")
            return result

    def to_dict(self):
        self._logger.debug("to_dict - waiting for lock")
        with self._tree_lock:
            self._update_tree()
            result = self.tree.to_dict()
            self._logger.info("to_dict success")
            return result

    def _build_tree(self):
        self._logger.debug("_build_tree - start")
        leafs = {}
        for gid, shadow in self._agentshadows.items():
            try:
                leafs[gid] = Leaf(shadow)
                self._logger.debug("_build_tree - added gid {}".format(gid))
            except ValueError:
                self._logger.debug("_build_tree - skipped gid {}".format(gid))
                pass
        self._logger.debug("_build_tree - creating root node with leafs {}".format(leafs))
        self.tree = Root(leafs)
        self._logger.debug("_build_tree - finish")

    def start(self):
        pass

    def stop(self):
        pass



def get_schema():
    schema = {
        "type": "object",
        "description": "all timings for the monitoring service",
        "properties": {
            "onboarding_timeout": {
                "type": "integer",
                "example": "60",
                "description": "seconds * 2 to be waited until an onboarding request should have been processed and a response been recevied"
            },
            "wait_for_message_timeout": {
                "type": "integer",
                "example": "180",
                "description": "seconds - if no message has been received for this period, the service is considered to be inactive"
            },
            "deactivation_timeout": {
                "type": "integer",
                "example": "360",
                "description": "seconds - an inactive service becomes a stopped service after this time span"
            },
            "heartbeat_interval": {
                "type": "integer",
                "example": "60",
                "description": "seconds - send heartbeat to all connected services"
            },
            "archivation_timeout": {
                "type": "integer",
                "example": "86400",
                "description": "seconds - a stopped service is archived after this time span."
            }
        },
        "required": ["onboarding_timeout", "wait_for_message_timeout", "deactivation_timeout",
                     "heartbeat_interval", "archivation_timeout"],
        "additionalProperties": False
    }
    return schema



import collections
import datetime
import pprint
from pelops.logging.mylogger import get_log_level
from pelops.logging.mylogger import get_level_name
import logging


class _ArchiveEntry:
    log = None
    counter = None
    time = None

    def __init__(self, maxlen):
        self.log = collections.deque(maxlen=maxlen)
        self.counter = 0

    def add(self, message):
        self.log.appendleft(message)
        self.counter += 1
        self.time = datetime.datetime.now()

    def get_stats(self):
        try:
            at = self.time.strftime("%Y-%m-%dT%H:%M:%S")
        except AttributeError:
            at = None
        return {
                   "received": self.counter,
                   "stored": len(self.log),
                   "time": at
               }

class LogArchive:
    flat_list = None
    level_dict = None
    max_level = None
    set_health = None
    _observer = None

    def __init__(self, maxlen=50, set_health=None):
        self.set_health = set_health
        self.reset_max_level()
        self.flat_list = _ArchiveEntry(maxlen)
        self.level_dict = collections.defaultdict(lambda: _ArchiveEntry(maxlen))

    def register_log_event_observer(self, observer):
        self._observer = observer

    def add(self, message):
        self.flat_list.add(message)
        level = get_log_level(message["level"])
        self.level_dict[level].add(message)
        self.max_level = max(self.max_level, level)
        self.set_health(self.max_level)
        if self._observer:
            self._observer(message["gid"], message["level"])

    def reset_max_level(self):
        self.max_level = logging.NOTSET
        self.set_health(self.max_level)

    def get_stats(self, full=True):
        stats = {
            "max_level": self.max_level,
            "logs": {
                "flat_list": {
                    "stats": self.flat_list.get_stats(),
                }
            }
        }
        if full:
            stats["logs"]["flat_list"]["entries"] = self.flat_list.log

        levels = [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        for k in levels:
            stats["logs"][get_level_name(k)] = {
                "stats": self.level_dict[k].get_stats()
            }
            if full:
                stats["logs"][get_level_name(k)]["entries"] = self.level_dict[k].log

        return stats

    def get_string_stats(self, key):
        output = ""
        
        if key.lower() == "flat":
            key = "flat_list"
        elif key.lower() == "debug":
            key = get_level_name(logging.DEBUG)
        elif key.lower() == "info":
            key = get_level_name(logging.INFO)
        elif key.lower() == "warning":
            key = get_level_name(logging.WARNING)
        elif key.lower() == "error":
            key = get_level_name(logging.ERROR)
        elif key.lower() == "critical":
            key = get_level_name(logging.CRITICAL)
        else:
            raise ValueError("get_string_stats - unknown argument value '{}'".format(key))

        logs = self.get_stats(full=True)["logs"]

        output += "statistics:\n{}\n\n".format(pprint.pformat(logs[key]["stats"]))
        output += "entries:\n{}\n\n".format(pprint.pformat(logs[key]["entries"]))
        
        return output



import pprint
import json
from hippodamia.flask.serverthread import ServerThread
from flask import Flask, render_template, url_for, flash
from flask_basicauth import BasicAuth
from flask.helpers import locked_cached_property
import logging
from pelops.logging.mylogger import get_child
from pelops.myconfigtools import mask_entries, dict_deepcopy_lowercase
import datetime


class LoggerFlask(Flask):
    _logger = None

    @staticmethod
    def _extract_handlers(logger):
        if logger.parent:
            handlers = LoggerFlask._extract_handlers(logger.parent)
        else:
            handlers = []
        handlers.extend(logger.handlers)
        return handlers

    def __init__(self, import_name, logger):
        self._logger = get_child(logger, "Flask")
        self._logger.setLevel(logger.level)
        werkzeug_logger = logging.getLogger('werkzeug')
        werkzeug_logger.setLevel(self._logger.level)
        Flask.__init__(self, import_name)

    @locked_cached_property
    def logger(self):
        return self._logger


class FlaskServer:
    _app = None
    _config = None
    _basic_auth = None
    _server_thread = None
    _port = None
    _logger = None

    _shadows = None
    _archive = None
    _hierarchical_view = None
    _heartbeat = None
    _onboarding = None
    _full_config = None
    _mqtt_client = None
    _request_ping = None
    _request_runtime = None
    _request_config = None
    _request_reonboarding = None
    _request_end = None
    _acknowledge_log = None
    _notification = None

    _archive_agentshadows = None
    _purge_agentshadowarchive = None

    _SECRET_KEY = '\xad\xb5\xac\xe1\x99\x1e\r\x9a\x8c\xc8\x07\x11\x06q\xe5\xbe\t\x19Na\xac\xd8\xa9\xe3'

    def __init__(self, config, logger, shadows, archive, heartbeat, hierarchical_view, onboarding, request_ping,
                 request_runtime, request_config, request_reonboarding, request_end, full_config, mqtt_client,
                 archive_agentshadows, purge_agentshadowarchive, acknowledge_log, notification):
        self._config = config
        self._logger = get_child(logger, self.__class__.__name__, self._config)
#        self._logger.addHandler(default_handler)
        self._logger.info("__init__ - start")
        self._logger.debug("__init__ - config: {}".format(self._config))

        self._shadows = shadows
        self._archive = archive

        self._hierarchical_view = hierarchical_view
        self._heartbeat = heartbeat
        self._onboarding = onboarding
        self._request_ping = request_ping
        self._request_runtime = request_runtime
        self._request_config = request_config
        self._request_reonboarding = request_reonboarding
        self._request_end = request_end
        self._acknowledge_log = acknowledge_log
        self._archive_agentshadows = archive_agentshadows
        self._purge_agentshadowarchive = purge_agentshadowarchive
        self._notification = notification

        self._full_config = full_config
        self._mqtt_client = mqtt_client

        self._port = self._config["port"]

        self._init_flask()

        self._logger.info("__init__ - finished")

    def _init_flask(self):
        self._app = LoggerFlask(__name__, self._logger)
        self._app.secret_key = self._SECRET_KEY
        self._basic_auth = self._create_basic_auth()
        self._add_routes()
        self._add_pprint_filter()
        self._add_psorteddict_filter()

    def _add_pprint_filter(self):
        self._logger.debug("add_pprint_filter to jinja")

        def to_pprint(value):
            if value is not None:
                return pprint.pformat(value, indent=2)
            return ""

        self._app.jinja_env.filters['pprint'] = to_pprint

    def _add_psorteddict_filter(self):
        self._logger.debug("add_psorted_filter to jinja")

        def to_psorted(value):
            text = ""
            if value is not None:
                if isinstance(value, dict):
                    keys = list(value.keys())
                    keys.sort()
                    for k in keys:
                        v = value[k]
                        text += "  '{}': {},\n".format(k, pprint.pformat(v, indent=4))
                    if len(text) == 0:
                        text = "{}"
                    else:
                        text = "{"+text[1:-2]+"}"
                elif isinstance(value, list):
                    value.sort()
                    text = pprint.pformat(value, indent=2)
                else:
                    text = str(value)
            return text

        self._app.jinja_env.filters["psorted"] = to_psorted

    def _create_basic_auth(self):
        if "webserver-user" in self._config and "webserver-password" in self._config:
            self._logger.info("_create_basic_auth - adding basic_auth")
            self._app.config['BASIC_AUTH_USERNAME'] = self._config["webserver-user"]
            self._app.config['BASIC_AUTH_PASSWORD'] = self._config["webserver-password"]
            self._app.config['BASIC_AUTH_FORCE'] = True
            return BasicAuth(self._app)
        else:
            self._logger.info("_create_basic_auth - skipping")
            return None

    def _run_server(self):
        self._logger.debug("app.run")
        self._app.run(port=self._port)

    def start(self):
        self._logger.info("starting flask server in process")
        self._server_thread = ServerThread(self._app, self._port)
        self._server_thread.start()
        self._logger.info("flask server started")

    def stop(self):
        self._logger.info("terminating flask server process")
        self._server_thread.stop()
        self._server_thread.join()
        self._logger.info("flask server stopped")

    def _add_routes(self):
        self._logger.debug("adding routes")
        self._app.add_url_rule('/', view_func=self.route_index)
        self._app.add_url_rule('/hierarchical', view_func=self.route_service_hierarchical)
        self._app.add_url_rule('/hierarchical/data', view_func=self.route_service_hierarchical_data)
        self._app.add_url_rule('/list', view_func=self.route_service_list)
        self._app.add_url_rule('/list/data', view_func=self.route_service_list_data)
        self._app.add_url_rule('/service/<gid>/commands', view_func=self.route_service_commands)
        self._app.add_url_rule('/service/<gid>/commands/<cmd>', view_func=self.execute_commands)
        self._app.add_url_rule('/service/<gid>/details', view_func=self.route_service_details)
        self._app.add_url_rule('/service/<gid>/messages', view_func=self.route_service_messages)
        self._app.add_url_rule('/service/<gid>/logs', view_func=self.route_service_logs)
        self._app.add_url_rule('/heartbeatstats', view_func=self.route_heartbeatstats)
        self._app.add_url_rule('/onboardingstats', view_func=self.route_onboardingstats)
        self._app.add_url_rule('/mqtt_stats', view_func=self.route_mqtt_stats)
        self._app.add_url_rule('/notificationstats', view_func=self.route_notificationstats)
        self._app.add_url_rule('/show_config', view_func=self.route_show_config)
        self._app.add_url_rule('/archive', view_func=self.route_archive_list)
        self._app.add_url_rule('/archive/data', view_func=self.route_archive_list_data)
        self._app.add_url_rule('/archive/<gid>/details', view_func=self.route_archive_details)
        self._app.add_url_rule('/archive/<gid>/messages', view_func=self.route_archive_messages)
        self._app.add_url_rule('/archive/<gid>/logs', view_func=self.route_archive_logs)
        self._app.add_url_rule('/group_commands', view_func=self.route_group_commands)
        self._app.add_url_rule('/group_commands/<cmd>', view_func=self.execute_group_commands)
        self._app.add_url_rule('/general_health', view_func=self.route_general_health)
        self._app.add_url_rule('/general_health_data', view_func=self.route_general_health_data)

    def route_index(self):
        return render_template('index.html', name="Monitoring Service")

    def route_general_health(self):
        return render_template('general_health.html')

    def route_general_health_data(self):
        health = self._hierarchical_view.to_dict()
        data = {
            "health": health["health"],
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }
        return json.dumps(data)

    def route_service_hierarchical(self):
        # output = self._hierarchical_view.pformat(details=ViewDetails.FULL)
        # title = "Hierarchical View"
        # return render_template('pre.html', title=title, content=output)
        title = "Hierarchical View"
        return render_template('hierarchical_view.html', title=title)

    def route_service_hierarchical_data(self):
        health = self._hierarchical_view.to_dict()
        for location in health["children"]:
            for room in location["children"]:
                for device in room["children"]:
                    for service in device["children"]:
                        gid = service["gid"]
                        service["route_service_details"] = url_for('route_service_details', gid=gid)
                        service["route_service_logs"] = url_for('route_service_logs', gid=gid)
                        service["route_service_commands"] = url_for('route_service_commands', gid=gid)
        data = {
            "health": health,
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }
        return json.dumps(data)

    def route_service_list(self):
        return render_template('service_list.html', title="Active Services",
                               update_route=url_for("route_service_list_data"))

    def route_service_list_data(self):
        data = {
            "entries": [],
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }

        for gid, shadow in self._shadows.items():
            entry = {
                "gid": gid,
                "name": shadow.properties.name,
                "type": shadow.properties.type,
                "state": shadow.get_state_id().name,
                "health": shadow.properties.health.name,
                "necessity": shadow.properties.necessity.name,
                "log_count": shadow.log_archive.flat_list.counter,
                "details": url_for('route_service_details', gid=gid),
                "messages": url_for('route_service_messages', gid=gid),
                "logs": url_for('route_service_logs', gid=gid),
                "commands": url_for('route_service_commands', gid=gid)
            }
            data["entries"].append(entry)
        return json.dumps(data)

    def route_service_commands(self, gid):
        if gid not in self._shadows:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_service_list_data"))

        return render_template('service_commands.html', title="Commands for '{}'".format(gid), gid=gid)

    def execute_commands(self, cmd, gid):
        if gid not in self._shadows:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_service_list_data"))

        message = "executed command '{}'".format(cmd)
        cmd = cmd.lower()
        if cmd == "request_ping":
            self._request_ping(gid)
        elif cmd == "request_runtime":
            self._request_runtime(gid)
        elif cmd == "request_config":
            self._request_config(gid)
        elif cmd == "request_reonboarding":
            self._request_reonboarding(gid)
        elif cmd == "request_end":
            self._request_end(gid)
        elif cmd == "acknowledge_log":
            counter = self._acknowledge_log(gid)
            if counter == 0:
                flash("acknowledge_log for gid '{}' failed.".format(gid))
            else:
                flash("acknowledge_log for gid '{}' success.".format(gid))
        else:
            message = "unknown command '{}'".format(cmd)

        return render_template('service_commands.html', title="Commands for '{}'".format(gid), gid=gid, message=message)

    def route_service_details(self, gid):
        try:
            stats = self._shadows[gid].get_stats()
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_service_list_data"))

        return render_template('service_details.html', title="Details for '{}'".format(gid), gid=gid, stats=stats)

    def route_service_logs(self, gid):
        try:
            log_entries = self._shadows[gid].log_archive.get_stats()["logs"]
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_service_list_data"))

        log_ids = ["flat_list", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

        logs = {}
        for k, shadow_logger in log_entries.items():
            logs[k] = {
                "counter_received": shadow_logger["stats"]["received"],
                "counter_stored": shadow_logger["stats"]["stored"],
                "counter_removed": shadow_logger["stats"]["received"] - shadow_logger["stats"]["stored"],
                "entries": shadow_logger["entries"],
            }

        return render_template('service_logs.html', title="Logs for '{}'".
                               format(gid), gid=gid, log_ids=log_ids, logs=logs)

    def route_service_messages(self, gid):
        try:
            messages = self._shadows[gid].message_archive.get_stats()
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_service_list_data"))

        message_types = list(messages.keys())
        message_types.sort()

        return render_template('service_messages.html', title="Messages for '{}'".
                               format(gid), gid=gid, messages=messages, message_types=message_types)

    def route_mqtt_stats(self):
        mqtt_stats = self._mqtt_client.get_stats()
        title = "MQTT Statistics"
        sorted_active_topics = list(mqtt_stats["subscribed_topics"])
        sorted_active_topics.sort()
        sorted_sendrecv_topics = list(mqtt_stats["mqtt_stats"]["topics"])
        sorted_sendrecv_topics.sort()
        return render_template('mqtt_stats.html', title=title, mqtt_stats=mqtt_stats,
                               sorted_active_topics=sorted_active_topics,
                               sorted_sendrecv_topics=sorted_sendrecv_topics)

    def route_heartbeatstats(self):
        stats = self._heartbeat.get_stats()
        title = "Heartbeat Statistics"
        return render_template('heartbeat.html', title=title, stats=stats)

    def route_onboardingstats(self):
        stats = self._onboarding.get_stats()
        title = "Onboarding Statistics"
        return render_template('onboarding.html', title=title, stats=stats)

    def route_notificationstats(self):
        if self._notification:
            stats = pprint.pformat(self._notification.get_stats(), indent=2)
        else:
            stats = ""
        title = "E-Mail Notification Statistics"
        return render_template('notification.html', title=title, stats=stats)

    def route_show_config(self):
        title = "Config"
        config = dict_deepcopy_lowercase(self._full_config)
        mask_entries(config)
        return render_template('config.html', title=title, config=config)

    def route_archive_list(self):
        return render_template('service_list.html', title="Service Archive",
                               update_route=url_for("route_archive_list_data"), archive=True)

    def route_archive_list_data(self):
        data = {
            "entries": [],
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }

        for shadow in self._archive:
            gid = shadow.properties.gid
            entry = {
                "gid": gid,
                "name": shadow.properties.name,
                "type": shadow.properties.type,
                "state": shadow.get_state_id().name,
                "health": shadow.properties.health.name,
                "necessity": shadow.properties.necessity.name,
                "log_count": shadow.log_archive.flat_list.counter,
                "details": url_for('route_archive_details', gid=gid),
                "messages": url_for('route_archive_messages', gid=gid),
                "logs": url_for('route_archive_logs', gid=gid),
                "commands": ""
            }
            data["entries"].append(entry)
        return json.dumps(data)

    def _get_archive_entry(self, gid):
        for entry in self._archive:
            if entry.properties.gid == gid:
                return entry
        raise KeyError("gid '{}' not found archive of agentshadows")

    def route_archive_details(self, gid):
        try:
            stats = self._get_archive_entry(gid).get_stats()
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_archive_list_data"))

        return render_template('service_details.html', title="Details for '{}'".format(gid), gid=gid, stats=stats,
                               archive=True)

    def route_archive_logs(self, gid):
        try:
            log_entries = self._get_archive_entry(gid).log_archive.get_stats()["log_entries"]
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_archive_list_data"))

        log_ids = ["flat_list", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

        logs = {}
        for k, shadow_logger in log_entries.items():
            logs[k] = {
                "counter_received": shadow_logger.counter,
                "counter_stored": len(shadow_logger.log),
                "counter_removed": shadow_logger.counter - len(shadow_logger.log),
                "entries": shadow_logger.log,
            }

        return render_template('service_logs.html', title="Logs for '{}'".
                               format(gid), gid=gid, log_ids=log_ids, logs=logs, archive=True)

    def route_archive_messages(self, gid):
        try:
            messages = self._get_archive_entry(gid).message_archive.get_stats()
        except KeyError:
            flash("unknown gid '{}'".format(gid))
            return render_template('service_list.html', title="Active Services",
                                   update_route=url_for("route_archive_list_data"))

        message_types = list(messages.keys())
        message_types.sort()

        return render_template('service_messages.html', title="Messages for '{}'".
                               format(gid), gid=gid, messages=messages, message_types=message_types, archive=True)

    def route_group_commands(self):
        return render_template('group_commands.html')

    def execute_group_commands(self, cmd):
        message = "executed group command '{}'".format(cmd)
        cmd = cmd.lower()
        if cmd == "request_ping":
            self._request_ping()
        elif cmd == "request_runtime":
            self._request_runtime()
        elif cmd == "request_config":
            self._request_config()
        elif cmd == "request_reonboarding":
            self._request_reonboarding()
        elif cmd == "request_end":
            self._request_end()
        elif cmd == "archive_agentshadows":
            result = self._archive_agentshadows()
            flash("archived {} agent shadows".format(result))
        elif cmd == "purge_agentshadowarchive":
            result = self._purge_agentshadowarchive()
            flash("purged {} from the agent shadow archive".format(result))
        elif cmd == "acknowledge_log":
            result = self._acknowledge_log()
            flash("acknowledged logs in {} agent shadows".format(result))
        else:
            message = "unknown command '{}'".format(cmd)

        return render_template('group_commands.html', message=message)





def get_schema():
    schema = {
        "type": "object",
        "description": "all topics for the monitoring services",
        "properties": {
            "onboarding-request": {
                "type": "string",
                "example": "/test/hippodamia/onboarding/request",
                "description": "hippodamia listens on this topic for new onboarding request"
            },
            "activity-base": {
                "type": "string",
                "example": "/test/hippodamia/incoming/",
                "description": "this topic will be used as base for each registered service. e.g .../incoming/1/ping"
            },
            "command-end": {
                "type": "string",
                "example": "/test/hippodamia/command/end",
                "description": "end command messages are published to this topic"
            },
            "command-reonboarding": {
                "type": "string",
                "example": "/test/hippodamia/command/reonboarding",
                "description": "reonboarding request are published to this topic"
            },
            "command-ping-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/ping_on_request",
                "description": "spontaneous ping messages are requested via this topic"
            },
            "command-config-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/command/config_on_request",
                "description": "spontaneous config messages are requested via this topic"
            },
            "command-runtime-on-request": {
                "type": "string",
                "example": "/test/hippodamia/command/command/runtime_on_request",
                "description": "spontaneous runtime messages are requested via this topic"
            },
            "heartbeat": {
                "type": "string",
                "example": "/test/hippodamia/heartbeat",
                "description": "heartbeat is published via this topic"
            }
        },
        "required": ["onboarding-request", "activity-base", "command-end", "command-reonboarding", "command-ping-on-request",
                     "command-config-on-request", "command-runtime-on-request", "heartbeat"],
        "additionalProperties": False
    }
    return schema



import datetime
import pprint
from hippodamia.enums import Enforcement
from hippodamia.agentshadow.states.state_ids import state_ids
import json
import pelops.logging.mylogger
from threading import Lock
import collections


class _GIDMapping:
    location = None
    room = None
    device = None
    type = None
    name = None

    def __init__(self, location=None, room=None, device=None, type=None, name=None):
        self.location = location
        self.room = room
        self.device = device
        self.type = type
        self.name = name

    def complete(self):
        return (self.location is not None
                and self.room is not None
                and self.device is not None
                and self.type is not None
                and self.name is not None)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.location == other.location \
               and self.room == other.room \
               and self.device == other.device \
               and self.type == other.type and self.name == other.name


class Onboarding:
    _logger = None
    _mqtt_client = None
    _agentshadows = None
    _asfactory = None
    _lock = None

    enforcement = None
    protocol_version = None
    topic_onboarding_request = None
    counter = 0
    gid_list = None
    last_timestamp = None
    messages = None
    gid_blacklist = None

    def __init__(self, protocol_version, topic_onboarding_request, enforcement, agentshadows, asfactory, mqtt_client, logger):
        self.protocol_version = protocol_version
        self._logger = pelops.logging.mylogger.get_child(logger, __class__.__name__)
        self._mqtt_client = mqtt_client
        self.enforcement = enforcement
        self._agentshadows = agentshadows
        self._asfactory = asfactory
        self.gid_list = {}
        self.gid_blacklist = []
        self.topic_onboarding_request = topic_onboarding_request
        self._lock = Lock()
        self.messages = collections.deque(maxlen=50)
        self._logger.info("__init__ done")

    def start(self):
        self._mqtt_client.subscribe(self.topic_onboarding_request, self._handler_onboarding_request)
        self._logger.info("subscribed to {}".format(self.topic_onboarding_request))

    def stop(self):
        self._mqtt_client.unsubscribe(self.topic_onboarding_request, self._handler_onboarding_request)
        self._logger.info("unsubscribed from {}".format(self.topic_onboarding_request))

    def _generate_unique_id(self, gid_mapping):
        if gid_mapping.type is None or gid_mapping.type == "":
            raise ValueError("value of gid_mapping.type must be set.")
        suffix = 0
        while True:
            new_gid = "{}-{}".format(gid_mapping.type, suffix)
            if new_gid not in self.gid_list.keys():
                return new_gid
            suffix += 1

    def _last_gid_ok(self, last_gid, gid_mapping):
        if last_gid is None or last_gid == "":
            # last_gid is not a valid identifier
            return False
        if last_gid in self.gid_list:
            last_mapping = self.gid_list[last_gid]
            if gid_mapping == last_mapping:
                # last_gid's mapping values are identical to the one from the requesting service
                return True
            else:
                # last_gid in use by a different microservice
                return False
        # last_gid is unknown
        return True

    def _get_gid_for_gidmapping(self, gid_mapping):
        for gid, gm in self.gid_list.items():
            if gm == gid_mapping:
                return gid
        return None

    def _is_known_gid_stopped(self, gid):
        try:
            state_id = self._agentshadows[gid].get_state_id()
            if not(state_id == state_ids.STOPPED or state_id == state_ids.ARCHIVED):
                return False
        except KeyError:
            pass
        return True

    def _same_session_id(self, gid, last_session):
        try:
            session = self._agentshadows[gid].properties.session
            if session is None or session == last_session:
                return True
        except KeyError:
            pass
        return False

    def _get_gid(self, identifier):
        gid_mapping = _GIDMapping(location=identifier["location"], room=identifier["room"], device=identifier["device"],
                                  type=identifier["type"], name=identifier["name"])
        gid = None
        self._logger.debug("_get_gid - gid_mapping: location={}, room={}, device={}, type={}, name={}; complete={}"
                           .format(gid_mapping.location, gid_mapping.room, gid_mapping.device, gid_mapping.type,
                                   gid_mapping.name, gid_mapping.complete()))

        if gid_mapping.complete():
            if self._last_gid_ok(identifier["last-gid"], gid_mapping):
                self._logger.debug("_get_gid - _last_gid_ok('{}') == True".format(identifier["last-gid"]))
                gid = identifier["last-gid"]
            else:
                gid = self._get_gid_for_gidmapping(gid_mapping)
                self._logger.debug("_get_gid - gid for gidmapping={}".format(gid))

        if not self._is_known_gid_stopped(gid) and not self._same_session_id(gid, identifier["last-session"]):
            self._logger.debug("_get_gid - agentshadow for gid {} is active (state: {}) and session ids do not match "
                               "(last-session: {})"
                               .format(gid, self._agentshadows[gid].get_state_id(), identifier["last-session"]))
            gid = None

        if gid in self.gid_blacklist:
            self._logger.debug("_get_gid - gid {} is blacklisted".format(gid))
            gid = None

        if gid is None:
            gid = self._generate_unique_id(gid_mapping)
            self._logger.debug("_get_gid - generated unique id")
            self.gid_list[gid] = gid_mapping

        self._logger.info("_get_gid - gid: {}".format(gid))
        return gid

    def _handler_onboarding_request(self, message):
        """
        {
            "uuid": "550e8400-e29b-11d4-a716-446655440000",
            "onboarding-topic": "/hippodamia/550e8400-e29b-11d4-a716-446655440000",
            "protocol-version": 1,
            "timestamp": "1985-04-12T23:20:50.520Z",
            "identifier": {
                "last-gid": "copreus-1",
                "type": "DriverManager",
                "module": "copreus.drivermanager",
                "name": "display-driver",
                "location": "flat",
                "room": "living room",
                "device": "thermostat",
                "decription": "lorem ipsum",
                "host-name": "rpi",
                "node-id": "00-07-E9-AB-CD-EF",
                "ips": [
                    "192.168.0.1",
                    "10.0.1.2",
                    "2001:0db8:85a3:08d3:1319:8a2e:0370:7344"
                ],
                "config-hash": "cf23df2207d99a74fbe169e3eba035e633b65d94"
            }
        }
        """
        with self._lock:
            message = message.decode("utf8")
            message = json.loads(message)
            self._logger.info("_handler_onboarding_request - received request")
            self._logger.debug("_handler_onboarding_request - message: {}".format(message))

            self.counter += 1
            self.last_timestamp = datetime.datetime.now()
            self.messages.append({"time": self.last_timestamp, "message": message})

            if message["protocol-version"] != self.protocol_version:
                message = "handler_onboarding_request - expected protocol version {}, message: {}"\
                    .format(self.protocol_version, message)
                self._logger.error(message)
            else:
                identifier = message["identifier"]
                self._logger.debug("_handler_onboarding_request - identifier: {}".format(identifier))
                gid = self._get_gid(identifier)

                shadow = None
                if gid not in self._agentshadows:
                    self._logger.debug("_handler_onboarding_request - unknown gid")
                    if self.enforcement == Enforcement.NONE:
                        self._logger.debug("_handler_onboarding_request - Enforcement.NONE")
                        shadow = self._asfactory.new_agentshadow(gid=gid)
                        self._agentshadows[gid] = shadow
                        shadow.start()
                    elif self.enforcement == Enforcement.IGNORE:
                        self._logger.debug("_handler_onboarding_request - Enforcement.IGNORE")
                        return
                    elif self.enforcement == Enforcement.STRICT:
                        self._logger.debug("_handler_onboarding_request - Enforcement.STRICT")
                        self._logger.error("handler_onboarding_request - required is set to strict and incoming is gid "
                                           "not pre-configured. '{}'".format(message))
                        return
                    else:
                        message = "handler_onboarding_request - dont know how to handler required '{}'"\
                            .format(self.enforcement)
                        self._logger.error(message)
                        raise ValueError(message)
                else:
                    self._logger.debug("_handler_onboarding_request - existing gid {}".format(gid))
                    shadow = self._agentshadows[gid]

                self._logger.debug("_handler_onboarding_request - {}".format(shadow))
                shadow.process_onboarding_request(message, session=self.counter)

    def get_stats(self, full=True):
        gid_list = list(self.gid_list.keys())
        gid_list.sort()

        stats = {
            "counter": self.counter,
            "stored": len(self.messages),
            "gid_list": gid_list,
            "enforcement": self.enforcement.name,
            "protocol_version": self.protocol_version,
            "topic_onboarding_request": self.topic_onboarding_request,
            "last_timestamp": self.last_timestamp
        }
        if full:
            stats["messages"] = self.messages
        return stats

    def get_string_stats(self):
        output = ""
        output += "stats:\n{}\n".format(pprint.pformat(self.get_stats(full=False), indent=4))
        try:
            output += "last message:\n{}\n".format(pprint.pformat(self.messages[-1], indent=4))
        except IndexError:
            output += "last message: n/a\n"

        return output



from hippodamia.enums import ViewDetails

"""
[ROOT]
  ├─ [ANode]
  │    ├─ [BNode]
  │    ├─ [CNode]
  │    │    ├─ [DNode]
  │    │    │    ├─ [ENode]
  │    │    │    └─ [FNode]
  │    │    └─ [GNode]
  │    └─ [HNode]
  └─ [INode]
       ├─ [JNode]
       ├─ [KNode]
       │    ├─ [LNode]
       │    └─ [MNode]
       └─ [NNode]


Root-ANode - prefix_first_line =  "  ├─ "
Root-ANode - prefix_next_lines =  "  │  "
ANode-CNode - prefix_first_line = "  │    ├─ "
ANode-CNode - prefix_next_lines = "  │    │  "
CNode-DNode - prefix_first_line = "  │    │    ├─ "
CNode-DNode - prefix_next_lines = "  │    │    │  "
INode-KNode - prefix_first_line = "       ├─ "
INode-KNode - prefix_next_lines = "       │  "

TYPE_A = "  ├─ "
TYPE_B = "  │  "
TYPE_C = "  └─ "
TYPE_D = "     "
"""


class Entry:
    MAX_NAME_LEN = 25
    INDENT = 2

    HORIZONTAL = "│"
    HORIZONTAL_CONNECTOR = "├"
    VERTICAL = "─"
    VERTICAL_CONNECTOR = "┬"
    EDGE = "└"
    SPACE = " "

    TYPE_A = SPACE + SPACE + HORIZONTAL_CONNECTOR + VERTICAL + SPACE
    TYPE_B = SPACE + SPACE + HORIZONTAL + SPACE + SPACE
    TYPE_C = SPACE + SPACE + EDGE + VERTICAL + SPACE
    TYPE_D = SPACE + SPACE + SPACE + SPACE + SPACE

    health = None
    name = None

    def __init__(self, name):
        if name is None or len(name) == 0:
            raise ValueError("Entry.__init__ - name must be set")
        self.name = name

    def update(self):
        raise NotImplementedError

    def _name2str(self):
        result = self.name
        if len(result) > self.MAX_NAME_LEN:
            result = result[:self.MAX_NAME_LEN - 3]
        result = "[{}]".format(result)
        return result

    def pformat(self, details=ViewDetails.NONE, filters=None):
        raise NotImplementedError

    def to_dict(self):
        raise NotImplementedError



from cmd import Cmd
import cmd
import datetime
from pelops.mythreading import LoggerThread
from hippodamia.enums import StatType
import pprint
from hippodamia.getch import getch


def _parse(arg):
    return tuple(map(str, arg.split()))


def more(text):
    _MAX_LINES = 35
    TEXT = "   --- Press ENTER to continue / ESC to abort ---"
    max_lines = _MAX_LINES
    lines = text.split("\n")
    i = 0
    for line in lines:
        print(line)
        i += 1
        if i == max_lines:
            i = 0
            print(TEXT, end='\r')
            while True:
                ch = getch()
                if ord(ch) == 27:
                    print(" " * len(TEXT), end='\r')
                    return  # ESC
                elif ord(ch) == 32:
                    max_lines = 1  # SPACE - single line feed
                    break
                elif ord(ch) == 13:
                    max_lines = _MAX_LINES  # ENTER - feed max lines
                    break
            print(" " * len(TEXT), end='\r')


class UI(Cmd):
    intro = None
    prompt = None

    _logger = None
    _shadows = None
    _health_view = None
    _request_ping = None
    _request_runtime = None
    _request_config = None
    _request_reonboarding = None
    _request_end = None
    _stop_service = None
    _heartbeat = None
    _onboarding = None
    _config = None

    _main_thread = None

    def __init__(self, version, config, shadows, heartbeat, health_view, onboarding, request_ping, request_runtime,
                 request_config, request_reonboarding, request_end, stop_service, logger):
        self.intro = "Hippodamia - Monitoring Service {}.   Type help or ? to list commands.\n".format(version)
        self.prompt = '(monitor) '

        Cmd.__init__(self)

        self._shadows = shadows
        self._config = config
        self._health_view = health_view
        self._request_ping = request_ping
        self._request_runtime = request_runtime
        self._request_config = request_config
        self._request_reonboarding = request_reonboarding
        self._request_end = request_end
        self._stop_service = stop_service
        self._logger = logger
        self._heartbeat = heartbeat
        self._onboarding = onboarding

        self._main_thread = LoggerThread(target=self._wrapper, name="ui.main", logger=self._logger)

    def start(self):
        self._main_thread.start()

    def stop(self):
        self._main_thread.join()

    def _wrapper(self):
        self.cmdloop()
        stop_daemon = LoggerThread(target=self._stop_service, name="ui.stop", logger=self._logger)
        stop_daemon.daemon = True
        stop_daemon.start()

    def _check_gid(self, gid):
        if gid not in self._shadows:
            raise KeyError("gid {} not in {}".format(gid, self._shadows.keys()))
        return gid

    def _request_extract_gid(self, name, arg):
        args = _parse(arg)

        gid = None
        if len(args) == 1 and str(args[0]).lower() == "all":
            print("Sending {} request to all monitored services\n".format(name))
        elif len(args) == 2 and str(args[0]).lower() == "gid":
            gid = self._check_gid(args[1])
            print("Send {} regest to monitoring service with gid == {}\n".format(name, gid))
        else:
            raise ValueError("Wrong arguments: {}. expected '{} [all|gid gid]'.\n".format(args, name))

        return gid

    def do_exit(self, arg):
        """exit - stops the monitoring service: EXIT"""
        return True

    def do_request_ping(self, arg):
        """request_ping [all|gid gid] - publish ping request either to all monitored services or to a specific one: REQUEST_PING GID 1"""
        try:
            self._request_ping(self._request_extract_gid("ping", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))

    def do_request_runtime(self, arg):
        """request_runtime [all|gid gid] - publish runtime request either to all monitored services or to a specific one: REQUEST_RUNTIME GID 1"""
        try:
            self._request_runtime(self._request_extract_gid("runtime", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_config(self, arg):
        """request_config [all|gid gid] - publish config request either to all monitored services or to a specific one: REQUEST_CONFIG GID 1"""
        try:
            self._request_config(self._request_extract_gid("config", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_reonboarding(self, arg):
        """request_reonboarding [all|gid gid] - publish reonboarding request either to all monitored services or to a specific one: REQUEST_REONBOARDING GID 1"""
        try:
            self._request_reonboarding(self._request_extract_gid("reonboarding", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))
            
    def do_request_end(self, arg):
        """request_end [all|gid gid] - publish end request either to all monitored services or to a specific one: REQUEST_END GID 1"""
        try:
            self._request_end(self._request_extract_gid("end", arg))
        except (KeyError, ValueError) as e:
            print("{}\n".format(e))

    def do_list(self, arg):
        """list - list all known services: LIST"""
        m = ""
        for gid, shadow in self._shadows.items():
            m += "gid: {}, state: {}, health:{}, necessity: {}, name: {}\n"\
                .format(gid, shadow.get_state_id().name, shadow.properties.health.name,
                        shadow.properties.necessity.name, shadow.properties.name)
        more(m)

    def do_hierarchical(self, arg):
        """"hierarchical - show hierarchical view of monitored services: HIERARCHICAL"""
        print("hierarchical view not implemented\n")

    def do_agentstats(self, arg):
        """agentstats gid gid [all|general|properties|ping|runtime|config|request|logger]
        show information on one specific monitored service: AGENTSTATS GID 1 PROPERTIES"""

        args = _parse(arg)
        if len(args) != 3:
            print("wrong number of arguments: {}\n".format(args))
        elif str(args[0]).lower() != "gid":
            print("first argument ({}) must be 'gid'\n".format(str(args[0]).lower()))
        elif str(args[2]).upper() not in StatType.get_members():
            print("third argument ({}) must be one of {}\n".format(str(args[2]).lower(), StatType.get_members()))
        else:
            try:
                stat_type = StatType.get_enum(str(args[2]))
                gid = self._check_gid(args[1])
                self._print_stats(gid, stat_type)
            except (ValueError, KeyError) as e:
                print("{}\n".format(e))

    def _print_stats(self, gid, type):
        print("stats {} for gid {}:\n".format(type, gid))

        shadow = self._shadows[gid]
        output = ""

        if type == StatType.GENERAL or type == StatType.ALL:
            output += "general: \n"
            output += " - active_state:\n"
            output += "   - name: {}\n".format(shadow.get_state_id().name)
            since = shadow.states[shadow.get_state_id()].timestamp_on_entry
            output += "   - since: {}\n".format(since)
            try:
                output += "   - duration: {}\n".format(datetime.datetime.now()-since)
            except TypeError:
                pass  # since is of type None
            output += " - machine state states: \n"
            for id, state in shadow.states.items():
                try:
                    duration = state.timestamp_on_exit - state.timestamp_on_entry
                except TypeError:
                    try:
                        duration = datetime.datetime.now() - state.timestamp_on_entry
                    except TypeError:
                        duration = "-"

                output += "   - {}: on_entry {} @{}, on_exit {} @{}, duration {}\n"\
                    .format(id.name, state.counter_on_entry, state.timestamp_on_entry,
                            state.counter_on_exit, state.timestamp_on_exit, duration)
            output += " - message stats: \n"
            for id, counter in shadow.shadow_data.get_stats().items():
                output += "   - {}: {}\n".format(id, counter)
            output += " - topics: \n"
            for id, topic in shadow.activity_topics.items():
                output += "   - {}: {}\n".format(id, topic)
            output += " - timings: \n"
            for id, timing in shadow.timings.items():
                output += "   - {}: {}\n".format(id, timing)

        if type == StatType.PROPERTIES or type == StatType.ALL:
            output += "properties: @{}\n".format(shadow.shadow_data.properties_time)
            output += pprint.pformat(shadow.shadow_data.properties, indent=4)
            output += "\n"

        if type == StatType.PING or type == StatType.ALL:
            output += "ping: #{} @{}\n".format(shadow.shadow_data.ping_counter, shadow.shadow_data.ping_time)
            output += pprint.pformat(shadow.shadow_data.ping, indent=4)
            output += "\n"

        if type == StatType.RUNTIME or type == StatType.ALL:
            output += "runtime: #{} @{}\n".format(shadow.shadow_data.runtime_counter, shadow.shadow_data.runtime_time)
            output += pprint.pformat(shadow.shadow_data.runtime, indent=4)
            output += "\n"

        if type == StatType.CONFIG or type == StatType.ALL:
            output += "config: #{} @{}\n".format(shadow.shadow_data.config_counter, shadow.shadow_data.config_time)
            output += pprint.pformat(shadow.shadow_data.config, indent=4)
            output += "\n"

        if type == StatType.REQUEST or type == StatType.ALL:
            output += "request: #{} @{}\n".format(shadow.shadow_data.request_counter, shadow.shadow_data.request_time)
            output += pprint.pformat(shadow.shadow_data.request, indent=4)
            output += "\n"

        if type == StatType.LOGGER or type == StatType.ALL:
            if len(shadow.shadow_data.logger) > 0:
                output += "logger: #{} @{}\n".format(shadow.shadow_data.logger_counter, shadow.shadow_data.logger_time)
                output += pprint.pformat(shadow.shadow_data.logger[-1], indent=4)
                output += "\n"
            else:
                output += "logger: ''\n"

        if output == "":
            print("UI._print_stats: unknown stat type '{}'\n".format(type))
        else:
            more(output)

    def do_agentlog(self, arg):
        """agentlog gid gid - view all received log-entries of this service: AGENTLOG GID 1"""

        args = _parse(arg)
        if len(args) != 2:
            print("wrong number of arguments: {}\n".format(args))
        elif str(args[0]).lower() != "gid":
            print("first argument ({}) must be 'gid'\n".format(str(args[0]).lower()))
        else:
            gid = self._check_gid(args[1])
            shadow = self._shadows[gid]
            print("show log for gid: {}\n".format(gid))

            output = "received: {}, stored: {}\n".format(shadow.shadow_data.logger_counter,
                                                         len(shadow.shadow_data.logger))
            counter = shadow.shadow_data.logger_counter
            for l in shadow.shadow_data.logger:
                output += "{}: {}\n".format(counter, pprint.pformat(l, indent=4))
                counter -= 1
            if counter > 0:
                output += "-- skipping {} removed log-entries".format(counter)
            more(output)

    def do_heartbeatstats(self, arg):
        """heartbeatstats - stats on the heartbeat signal of the monitoring agent: HEARTBEATSTATS"""
        output = "heartbeat: \n"
        output += " - counter: {}\n".format(self._heartbeat.counter)
        output += " - last timestamp: {}\n".format(self._heartbeat.last_timestamp)
        output += " - interval: {} s\n".format(self._heartbeat.interval)
        output += " - topic: {}\n".format(self._heartbeat.topic)
        more(output)

    def do_onboardingstats(self, arg):
        """onboardingstats - stats on the overall onboarding process: ONBOARDINGSTATS"""
        output = "onboarding: \n"
        output += " - counter: {}\n".format(self._onboarding.counter)
        output += " - gid_list: {}\n".format(self._onboarding.gid_list)
        output += " - enforcement: {}\n".format(self._onboarding.enforcement.name)
        output += " - protocol_version: {}\n".format(self._onboarding.protocol_version)
        output += " - topic_onboarding_request: {}\n".format(self._onboarding.topic_onboarding_request)
        output += " - last_timestamp: {}\n".format(self._onboarding.last_timestamp)
        output += " - last_message: \n"
        output += pprint.pformat(self._onboarding.last_message, indent=4)
        more(output)

    def do_config(self, arg):
        """config - show the current config: CONFIG"""
        output = pprint.pformat(self._config, indent=2)
        more(output)

    def do_a(self, arg):
        # """shortcut for 'stats gid 1 all'"""
        self._print_stats("1", StatType.ALL)

    def do_g(self, arg):
        # """shortcut for 'stats gid 1 general'"""
        self._print_stats("1", StatType.GENERAL)

    def do_e(self, arg):
        # """shortcut for 'exit'"""
        return True

    def do_q(self, arg):
        # """shortcut for 'exit'"""
        return True

    def do_h(self, arg):
        # """shortcut for 'help'"""
        self.do_help(arg)

    def do_l(self, arg):
        # """shortcut for 'list'"""
        self.do_list(arg)

    def do_o(self, arg):
        # """ shortcut for 'onboardingstats'"""
        self.do_onboardingstats(arg)



import datetime
import time
import pprint
from hippodamia.enums import ViewDetails
from hippodamia.enums import StatType
from hippodamia.email.anotifier import ANotifier


class Digest(ANotifier):
    _repeat = None
    _at = None
    _start_day = None
    _send_at_startup = None

    _agentshadows = None
    _heartbeat = None
    _hierarchical_view = None
    _onboarding = None
    _archive = None
    _mqtt_client = None

    def __init__(self, config, send_mail, logger, mqtt_client, agentshadows, heartbeat, hierarchical_view,
                 onboarding, archive):
        ANotifier.__init__(self, config, send_mail, logger)

        self._repeat = self._config["repeat"]
        try:
            temp = self._config["at"].split(":")
            self._at = datetime.time(hour=int(temp[0]), minute=int(temp[1]))
        except KeyError:
            self._at = datetime.datetime.now().time()
        self._logger.debug("__init__ - setting 'at' to {}".format(self._at))
        try:
            weekday = self._config["start-day"]
            weekdays = ["MONDAY", "TUESDAY", "WEDNESDAY", "THURSDAY", "FRIDAY", "SATURDAY", "SUNDAY"]
            self._start_day = weekdays.index(weekday)
        except KeyError:
            self._start_day = datetime.datetime.now().weekday()
        self._logger.debug("__init__ - setting 'start-day' to {}".format(self._start_day))
        try:
            self._send_at_startup = self._config["send-at-startup"]
        except:
            self._send_at_startup = False

        if self._send_at_startup and "start-day" not in self._config and "at" not in self._config:
            self._send_at_startup = False
            self._logger.debug("__init__ - forcing 'send-at-startup' to {} - 'start.day' and 'at' are not set -> a "
                               "digest email is sent immediately after startup anyways.".format(self._send_at_startup))
        else:
            self._logger.debug("__init__ - setting 'send-at-startup' to {}".format(self._send_at_startup))

        self._mqtt_client = mqtt_client
        self._agentshadows = agentshadows
        self._heartbeat = heartbeat
        self._hierarchical_view = hierarchical_view
        self._onboarding = onboarding
        self._archive = archive

    def _calc_first_timestamp(self):
        diff = self._start_day - datetime.datetime.now().weekday()
        if diff < 0:
            diff += 7
        days = datetime.timedelta(days=diff)
        target_day = datetime.datetime.now().date() + days
        target_time = self._at
        target = datetime.datetime.combine(target_day, target_time)
        first = time.mktime(target.timetuple())
        self._logger.info("_calc_first_timestamp: at {} seconds (now {} s)".format(first, time.time()))
        return first

    def _calc_interval(self):
        if self._repeat == "HOURLY":
            interval = 3600
        elif self._repeat == "DAILY":
            interval = 3600 * 24
        elif self._repeat == "WEEKLY":
            interval = 3600 * 24 * 7
        else:
            err_message = "_calc_interval - unknown repeat '{}'".format(self._repeat)
            self._logger.error(err_message)
            raise RuntimeError(err_message)

        self._logger.info("_calc_interval: {} seconds".format(interval))
        return interval

    def _digest_send_message(self):
        body = self._render_body()
        self._send_message(body)

    def _start(self):
        self._sched.repeatabs(self._calc_first_timestamp(), self._calc_interval(), 1, self._digest_send_message)
        if self._send_at_startup:
            self._logger.info("sending digest at startup")
            self._digest_send_message()

    def _stop(self):
        pass

    def _render_body(self):
        message  = "<h1>Hierarchical View</h1>"
        message += "<pre>{}</pre>".format(self._hierarchical_view.pformat(details=ViewDetails.FULL))

        message += "<h1>Micro-Service Overview</h1>"
        keys = list(self._agentshadows.keys())
        keys.sort()
        for k in keys:
            microservice = self._agentshadows[k]
            message += "<h2>[{}] {}</h2>".format(microservice.properties.gid, microservice.properties.name)
            message += "<pre>{}</pre>".format(microservice.get_string_stats(StatType.GENERAL))

        message += "<h1>Micro-Service Archive</h1>"
        if len(self._archive) == 0:
            message += "<div>Archive is empty.</div>"
        else:
            message = "<ul>"
            for microservice in self._archive:
                message += " <li>[{}] {}</li>".format(microservice.properties.gid, microservice.properties.name)
            message += "</ul>"

        message += "<h1>Heartbeat</h1>"
        message += "<pre>{}</pre>".format(self._heartbeat.get_string_stats())

        message += "<h1>Onboarding</h1>"
        message += "<pre>{}</pre>".format(self._onboarding.get_string_stats())

        message += "<h1>MQTT</h1>"
        message += "<div>mqtt client connection active: <div>{}</div>".format(self._mqtt_client.is_connected.is_set())
        message += "<h2>active subscriptions</h2>"
        message += "<ul>"
        for sub, func in self._mqtt_client.subscribed_topics().items():
            message += " - {}: [{}]\n".format(sub, func)
        message += "</ul>"
        message += "<h2>send/receive statistics</h2>"
        message += "<pre>{}</pre>".format(pprint.pformat(self._mqtt_client.stats.get_stats(), indent=2))

        return message

    def _message_footer(self):
        message = "<br/>"
        message += "<div style='text-align:right'>Digest generated at {}</div>".format(datetime.datetime.now())
        try:
            ne = self._sched.scheduler.queue[0]
            ne_time = ne.time
            dt = datetime.datetime.fromtimestamp(ne_time)
            next_event = dt.strftime("%Y-%m-%d %H:%M:%S.%f")
        except IndexError:
            next_event = "n/a"
        message += "<div style='text-align:right'>Next digest scheduled for {}</div>".format(next_event)
        return message

    def _get_stats(self):
        return {}



from hippodamia.enums import StatType
from collections import namedtuple
from collections import defaultdict
from collections import deque
from threading import Lock
from hippodamia.email.anotifier import ANotifier
import datetime
from hippodamia.agentshadow.states.state_ids import state_ids


class StateTransition:
    from_state = None
    to_state = None

    def __init__(self, from_state, to_state):
        self.from_state = from_state
        self.to_state = to_state

    def __eq__(self, other):
        if not isinstance(other, StateTransition):
            return NotImplemented
        return self.from_state == other.from_state and self.to_state == other.to_state

    def __hash__(self):
        return hash((self.from_state, self.to_state))

    def __repr__(self):
        return "StateTransition(from_state='{}', to_state='{}')".format(self.from_state, self.to_state)


class Event:
    gid = None
    type = None
    value = None
    time = None

    def __init__(self, gid, value):
        self.gid = gid
        self.value = value
        self.time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")

    def to_html(self):
        raise NotImplementedError


class StateEvent(Event):
    def __init__(self, gid, value):
        Event.__init__(self, gid, value)

    def to_html(self):
        html = "<span><b>{}</b> from: {}, to: {}, @{}</span>".format(self.gid, self.value.from_state, self.value.to_state,
                                                                   self.time)
        return html


class LogEvent(Event):
    def __init__(self, gid, value):
        Event.__init__(self, gid, value)

    def to_html(self):
        html = "<span><b>{}</b> log-level: {}, @{}</span>".format(self.gid, self.value, self.time)
        return html


class Spontaneous(ANotifier):
    _state_filter = None
    _event_queue = None
    _event_queue_lock = None
    _delay = None
    _state_transition_names = None
    _agentshadows = None

    def __init__(self, config, send_mail, logger, agentshadows, agentshadowfactory):
        ANotifier.__init__(self, config, send_mail, logger)

        self._agentshadows = agentshadows

        self._delay = self._config["delay"]

        self._state_transition_names = {}
        self._state_filter = []
        for cse in self._config["state_events"]:
            se = StateTransition(cse["from"], cse["to"])
            self._state_filter.append(se)
            se2 = StateTransition(state_ids.INACTIVE.name, state_ids.LOST.name)
            name = "{} -> {}".format(se.from_state, se.to_state)
            self._state_transition_names[se] = name
            self._logger.debug("__init__ - adding state filter '{}'".format(name))

        self._log_level_filter = []
        for l in self._config["log_events"]:
            self._log_level_filter.append(l)
            self._logger.debug("__init__ - adding log level filter '{}'".format(l))

        self._event_queue = deque()
        self._event_queue_lock = Lock()

        agentshadowfactory.register_state_transistion_event_observer(self.state_transition_event, agentshadows)
        agentshadowfactory.register_log_event_observer(self.log_event, agentshadows)

    def _start(self):
        pass

    def _stop(self):
        pass

    def _process_queue(self):
        with self._event_queue_lock:
            if self._event_queue:
                event_mapping = defaultdict(list)
                while self._event_queue:
                    event = self._event_queue.pop()
                    event_mapping[event.gid].append(event)
                body = self._render_body(event_mapping)
                self._send_message(body)

    def _render_body(self, event_mapping):
        text = "<h1>Affected Agent Shadows</h1>"
        text += "<ul>"
        keys = list(event_mapping.keys())
        keys.sort()
        for gid in keys:
            shadow = self._agentshadows[gid]
            text += "<li>[{}] {}</li>".format(gid, shadow.properties.name)
        text += "</ul>"

        for gid, mapping in event_mapping.items():
            shadow = self._agentshadows[gid]
            text += "<h1>[{}] {}</h1>".format(gid, shadow.properties.name)
            text += "<h2>events</h2>"
            text += "<ul>"
            for m in mapping:
                text += "<li>{}</li>".format(m.to_html())
            text += "</ul>"
            text += "<h2>agent shadow details</h2>"
            text += "<pre>{}</pre>".format(shadow.get_string_stats(StatType.ALL))

        return text

    def state_transition_event(self, gid, from_state, to_state):
        st = StateTransition(from_state.name, to_state.name)
        if st in self._state_filter:
            event = StateEvent(gid, st)
            self._put(event)

    def log_event(self, gid, level):
        if level in self._log_level_filter:
            event = LogEvent(gid, level)
            self._put(event)

    def _put(self, event):
        with self._event_queue_lock:
            self._event_queue.appendleft(event)
            self._sched.enter(self._delay, 1, self._process_queue)

    def _message_footer(self):
        message = "<br/>"
        message += "<div style='text-align:right'>Event notification generated at {}</div>".format(datetime.datetime.now())
        return message

    def _get_stats(self):
        return {}



import json
import pprint
import pelops.logging.mylogger
from hippodamia.enums import Necessity
from hippodamia.enums import StatType
import hippodamia.agentshadow.states.state_machine
from hippodamia.agentshadow.states.event_ids import event_ids
from hippodamia.agentshadow.states.state_ids import state_ids
from hippodamia.agentshadow.messagearchive import MessageArchive
from hippodamia.agentshadow.logarchive import LogArchive
from hippodamia.agentshadow.properties import Properties


class AgentShadow:
    _mqtt_client = None
    _logger = None

    _state_machine = None

    _basic_topics = None
    _topic_cmd_reonboarding = None

    service_timings = None
    _agent_timings = None
    activity_topics = None

    message_archive = None
    log_archive = None
    properties = None
    states = None

    def __init__(self, topics, service_timings, agent_timings, mongo_client, mqtt_client, logger,
                 preload_config=None, gid=None):
        self._mqtt_client = mqtt_client
        self._logger = pelops.logging.mylogger.get_child(logger, "AgentShadow")
        self._basic_topics = topics
        self._topic_cmd_reonboarding = self._basic_topics["command-reonboarding"]
        self.service_timings = service_timings
        self._agent_timings = agent_timings

        self.properties = Properties(self._logger)

        if gid:
            self._logger.info("__init__ - setting gid")
            self.properties.gid = str(gid)
            self.properties.necessity = Necessity.SPONTANEOUS
        elif preload_config:
            self._process_preload_config(preload_config)
        else:
            self._logger.error("__init__ - preload_config and gid are None")
            raise ValueError("__init__ - preload_config and gid are None")

        if self.properties.gid is None:
            self._logger.error("__init__ - gid is None")
            raise ValueError("__init__ - gid is None")
        else:
            self._logger = pelops.logging.mylogger.get_child(logger, str(self.properties.gid))

        self.message_archive = MessageArchive()
        self.log_archive = LogArchive(set_health=self.properties.set_log_health)

        self._set_activity_topics()
        self._init_state_machine((preload_config is not None))

    def get_state_id(self):
        return self._state_machine.get_active_state().id

    def _set_activity_topics(self):
        self.activity_topics = {}
        activity_base = self._basic_topics["activity-base"]
        if activity_base[-1] == "/":
            activity_base = activity_base[:-1]
        self.activity_topics["ping"] = "{}/{}/ping".format(activity_base, self.properties.gid)
        self.activity_topics["runtime"] = "{}/{}/runtime".format(activity_base, self.properties.gid)
        self.activity_topics["config"] = "{}/{}/config".format(activity_base, self.properties.gid)
        self.activity_topics["end"] = "{}/{}/end".format(activity_base, self.properties.gid)
        self.activity_topics["logger"] = "{}/{}/logger".format(activity_base, self.properties.gid)

    def _process_preload_config(self, preload_config):
        self._logger.info("_process_preload_config - start")
        self._logger.debug("_process_preload_config: {}".format(preload_config))

        if "gid" in preload_config:
            self.properties.gid = str(preload_config["gid"])
        if "name" in preload_config:
            self.properties.name = preload_config["name"]
        if "location" in preload_config:
            self.properties.location = preload_config["location"]
        if "room" in preload_config:
            self.properties.room = preload_config["room"]
        if "device" in preload_config:
            self.properties.device = preload_config["device"]
        if "necessity" in preload_config:
            necessity = Necessity.get_enum(preload_config["necessity"])
            if necessity != Necessity.OPTIONAL and necessity != Necessity.REQUIRED:
                msg = "_process_preload_config - unexpected value '{}' for necessity. must " \
                      "be 'required' or 'optional'.",format(preload_config["necessity"])
                self._logger.error(msg)
                raise ValueError(msg)
            self.properties.necessity = necessity
        self._logger.info("_process_preload_config - finished")

    def _init_state_machine(self, predefined):
        self._state_machine, self.states = hippodamia.agentshadow.states.\
            state_machine.create(self.properties.gid, self.service_timings,
                                 self.properties.set_state_health,
                                 self.log_archive.reset_max_level, self._logger, predefined=predefined)
        self.states[state_ids.ONBOARDING].func_send_onboarding_response = self._send_onboarding_response
        self.states[state_ids.ONBOARDING].func_deactivate_recv_messages = self.deactivate_recv_messages
        self.states[state_ids.ONBOARDING].func_activate_recv_messages = self.activate_recv_messages
        self.states[state_ids.STOPPED].func_deactivate_recv_messages = self.deactivate_recv_messages

    def start(self):
        self._state_machine.start()

    def stop(self):
        try:
            # trying to initiate offboarding no matter what the current state is
            self._state_machine.asyncoperate(event_ids.OFFBOARDING)
        except KeyError:
            pass  # not every every state has a link to offboarding
        self._state_machine.stop()

    def prepare_for_archivation(self):
        self._logger.info("prepare for archivation - start")
        self._state_machine.operate(event_ids.ARCHIVE)
        if self.get_state_id() != state_ids.ARCHIVED:
            e = RuntimeError("transition to state 'ARCHIVE' failed. current state: '{}'".
                             format(self.get_state_id().name))
            self._logger.exception(e)
            raise e
        self._state_machine.stop()
        self.deactivate_recv_messages()
        self._logger.info("prepare for archivation - finished")

    def register_state_transistion_event_observer(self, observer):
        self._logger.info("register_state_transistion_event_observer - registering observer '{}'".format(observer))
        self._state_machine.register_state_transistion_event_observer(observer)

    def register_log_event_observer(self, observer):
        self._logger.info("register_log_event_observer - registering observer '{}'".format(observer))
        self.log_archive.register_log_event_observer(observer)

    def process_onboarding_request(self, message, session):
        """
        {
          "uuid": "550e8400-e29b-11d4-a716-446655440000",
          "onboarding-topic": "/hippodamia/550e8400-e29b-11d4-a716-446655440000",
          "protocol-version": 1,
          "timestamp": "1985-04-12T23:20:50.520Z",
          "identifier": {
            "last-gid": 1,
            "type": "DriverManager",
            "module": "copreus.drivermanager",
            "version": "0.3.1",
            "name": "display-driver",
            "location": "flat",
            "room": "living room",
            "device": "thermostat",
            "decription": "lorem ipsum",
            "host-name": "rpi",
            "node-id": "00-07-E9-AB-CD-EF",
            "ips": [
              "192.168.0.1",
              "10.0.1.2",
              "2001:0db8:85a3:08d3:1319:8a2e:0370:7344"
            ],
            "config-hash": "cf23df2207d99a74fbe169e3eba035e633b65d94"
          }
        }
        """
        self._logger.info("process_onboarding_request - start session '{}'".format(session))
        self._logger.debug("process_onboarding_request - message: {}".format(message))

        self.properties.session = session
        self.properties.uuid = message["uuid"]
        self.properties.onboarding_topic = message["onboarding-topic"]
        self.properties.protocol_version = message["protocol-version"]
        self.properties.time = message["timestamp"]

        self.properties.location = message["identifier"]["location"]
        self.properties.room = message["identifier"]["room"]
        self.properties.device = message["identifier"]["device"]

        self.properties.type = message["identifier"]["type"]
        self.properties.module = message["identifier"]["module"]
        self.properties.version = message["identifier"]["version"]
        self.properties.name = message["identifier"]["name"]
        self.properties.description = message["identifier"]["description"]
        self.properties.host_name = message["identifier"]["host-name"]
        self.properties.node_id = message["identifier"]["node-id"]
        self.properties.ips = message["identifier"]["ips"]
        self.properties.config_hash = message["identifier"]["config-hash"]

        self.message_archive.archive_request(message)
        self._logger.info("process_onboarding_request - finished")
        self._state_machine.asyncoperate(event_ids.ONBOARDING_REQUEST)
        
    def _send_onboarding_response(self):
        """
        {
            "uuid": "550e8400-e29b-11d4-a716-446655440000",
            "gid": 1,
            "session": 123,
            "topics-activity": {
                "ping": "/hippodamia/commands",
                "runtime": "/hippodamia/commands",
                "config": "/hippodamia/commands",
                "end": "/hippodamia/commands",
                "logger": "/hippodamia/commands"
            },
            "topics-commands": {
                "end": "/hippodamia/commands",
                "reonboarding": "/hippodamia/commands",
                "ping_on_request": "/hippodamia/commands",
                "config_on_request": "/hippodamia/commands",
                "runtime_on_request": "/hippodamia/commands",
                "heartbeat": "/hippodamia/commands"
            },
            "timings": {
                "send-ping": 60,
                "send-runtime": 500,
                "send-config": 3600,
                "expect-heartbeat": 120
            }
        }
        """
        message = {
            "uuid": self.properties.uuid,
            "gid": self.properties.gid,
            "session": self.properties.session,
            "topics-activity": {
                "ping": self.activity_topics["ping"],
                "runtime": self.activity_topics["runtime"],
                "config": self.activity_topics["config"],
                "end": self.activity_topics["end"],
                "logger": self.activity_topics["logger"],
            },
            "topics-commands": {
                "end": self._basic_topics["command-end"],
                "reonboarding": self._basic_topics["command-reonboarding"],
                "ping_on_request": self._basic_topics["command-ping-on-request"],
                "config_on_request": self._basic_topics["command-config-on-request"],
                "runtime_on_request": self._basic_topics["command-runtime-on-request"],
                "heartbeat": self._basic_topics["heartbeat"],
            },
            "timings": {
                "send-ping": self._agent_timings["send-ping"],
                "send-runtime": self._agent_timings["send-runtime"],
                "send-config": self._agent_timings["send-config"],
                "expect-heartbeat": self._agent_timings["expect-heartbeat"]
            }
        }
        message = json.dumps(message)
        self._logger.debug("_send_onboarding_response - topic: {}, message: {}"
                           .format(self.properties.onboarding_topic, message))
        self._logger.debug(self.properties.get_stats())
        self._mqtt_client.publish(self.properties.onboarding_topic, message)

    def activate_recv_messages(self):
        self._logger.info("activate_recv_messages")

        self._logger.debug("activate_recv_messages - topic ping {}".format(self.activity_topics["ping"]))
        self._mqtt_client.subscribe(self.activity_topics["ping"], self._handler_ping, ignore_duplicate=True)

        self._logger.debug("activate_recv_messages - topic runtime {}".format(self.activity_topics["runtime"]))
        self._mqtt_client.subscribe(self.activity_topics["runtime"], self._handler_runtime, ignore_duplicate=True)

        self._logger.debug("activate_recv_messages - topic config {}".format(self.activity_topics["config"]))
        self._mqtt_client.subscribe(self.activity_topics["config"], self._handler_config, ignore_duplicate=True)

        self._logger.debug("activate_recv_messages - topic end {}".format(self.activity_topics["end"]))
        self._mqtt_client.subscribe(self.activity_topics["end"], self._handler_end, ignore_duplicate=True)

        self._logger.debug("activate_recv_messages - topic logger {}".format(self.activity_topics["logger"]))
        self._mqtt_client.subscribe(self.activity_topics["logger"], self._handler_logger, ignore_duplicate=True)

    def deactivate_recv_messages(self):
        self._logger.info("deactivate_recv_messages")

        self._logger.debug("activate_recv_messages - topic ping {}".format(self.activity_topics["ping"]))
        self._mqtt_client.unsubscribe(self.activity_topics["ping"], self._handler_ping, ignore_not_found=True)

        self._logger.debug("activate_recv_messages - topic runtime {}".format(self.activity_topics["runtime"]))
        self._mqtt_client.unsubscribe(self.activity_topics["runtime"], self._handler_runtime, ignore_not_found=True)

        self._logger.debug("activate_recv_messages - topic config {}".format(self.activity_topics["config"]))
        self._mqtt_client.unsubscribe(self.activity_topics["config"], self._handler_config, ignore_not_found=True)

        self._logger.debug("activate_recv_messages - topic end {}".format(self.activity_topics["end"]))
        self._mqtt_client.unsubscribe(self.activity_topics["end"], self._handler_end, ignore_not_found=True)

        self._logger.debug("activate_recv_messages - topic logger {}".format(self.activity_topics["logger"]))
        self._mqtt_client.unsubscribe(self.activity_topics["logger"], self._handler_logger, ignore_not_found=True)

    def _check_session(self, message):
        if self.properties.session != message["session"]:
            self._logger.warning("agentshadow.session != message.session ({} != {})"
                                 .format(self.properties.session, message["session"]))
            message = {
                "request": "reonboarding",
                "gid": self.properties.gid
            }
            payload = json.dumps(message)
            self._logger.info("_check_session - sending reonboarding request to gid '{}'".format(self.properties.gid))
            self._mqtt_client.publish(self._topic_cmd_reonboarding, payload)

    def _handler_ping(self, message):
        self._logger.info("_handler_ping - received ping")
        message = json.loads(message.decode("utf8"))
        self._check_session(message)
        self.message_archive.archive_ping(message)
        self.properties.service_uptime = message["service-uptime"]
        self._resettimer_or_regularmessage()

    def _handler_runtime(self, message):
        self._logger.info("_handler_runtime - received runtime")
        message = json.loads(message.decode("utf8"))
        self.message_archive.archive_runtime(message)
        self.properties.service_uptime = message["service-uptime"]
        self._resettimer_or_regularmessage()

    def _handler_config(self, message):
        self._logger.info("_handler_config - received config")
        message = json.loads(message.decode("utf8"))
        self.message_archive.archive_config(message)
        self.properties.service_uptime = message["service-uptime"]
        self._resettimer_or_regularmessage()

    def _handler_logger(self, message):
        self._logger.info("_handler_logger - received logger")
        message = json.loads(message.decode("utf8"))
        self.message_archive.archive_logger(message)
        self.log_archive.add(message)
        self._resettimer_or_regularmessage()

    def _resettimer_or_regularmessage(self):
        active_state = self._state_machine.get_active_state()
        if active_state.id == state_ids.ACTIVE:
            self._logger.info("current state is 'ACTIVE' -> reset timeout timer")
            self._state_machine.restarttimeoutevent()
        else:
            self._logger.info("current state is '{}' -> trigger event REGULAR_MESSAGE".format(active_state.id))
            self._state_machine.asyncoperate(event_ids.REGULAR_MESSAGE)

    def _handler_end(self, message):
        self._logger.info("_handler_end - received end")
        message = json.loads(message.decode("utf8"))
        self.message_archive.archive_end(message)
        self._state_machine.asyncoperate(event_ids.OFFBOARDING)

    def _get_general_stats(self):
        stats = dict()

        stats["micro-service"] = {
            "name": self.properties.name,
            "module": self.properties.module,
            "type": self.properties.type,
            "uptime": self.properties.service_uptime,
            "session": self.properties.session,
            "health": self.properties.health.name
        }

        stats["active-state"] = self._state_machine.get_stats_active_state()

        stats["message-stats"] = {}
        for id, gs in self.message_archive.get_stats().items():
            stats["message-stats"][id] = gs["counter"]

        stats["topics"] = {}
        for id, topic in self.activity_topics.items():
            stats["topics"][id] = topic

        stats["timings"] = {}
        for id, timing in self.service_timings.items():
            stats["timings"][id] = timing

        return stats

    def _get_state_stats(self):
        stats = dict()
        for k, v in self.states.items():
            stats[k.name] = v.get_stats()
        return stats

    def get_stats(self):
        stats = {
            "general": self._get_general_stats(),
            "states": self._get_state_stats(),
            "state_machine": self._state_machine.get_stats(),
            "properties": self.properties.get_stats(),
            "messages": self.message_archive.get_stats(full=False),
            "logger": self.log_archive.get_stats(full=False)
        }

        return stats

    def get_string_stats(self, stat_type):
        stats = self.get_stats()
        output = ""

        if stat_type == StatType.GENERAL or stat_type == StatType.ALL:
            output += "general: \n"
            output += pprint.pformat(stats["general"], indent=4)
            output += "\n\n"

        if stat_type == StatType.STATE_MACHINE or stat_type == StatType.ALL:
            output += "state-machine:\n"
            output += pprint.pformat(stats["state_machine"], indent=4)
            output += "\n\n"

        if stat_type == StatType.PROPERTIES or stat_type == StatType.ALL:
            output += "properties: @{}\n".format(stats["properties"]["time"])
            output += pprint.pformat(stats["properties"]["entries"], indent=4)
            output += "\n\n"

        if stat_type == StatType.LOGGER or stat_type == StatType.ALL:
            output += "logger:\n"
            output += pprint.pformat(stats["logger"], indent=4)
            output += "\n\n"

        if stat_type == StatType.MESSAGES or stat_type == StatType.ALL:
            output += "messages:\n"
            output += pprint.pformat(stats["messages"], indent=4)
            output += "\n\n"

        if output == "":
            output = "get_string_stats: unknown stat type '{}'\n".format(type)

        return output



import hippodamia.schema.topics as topics
import hippodamia.schema.service_timings as service_timings
import hippodamia.schema.agent_timings as agent_timings
import hippodamia.schema.expected_services as expected_services
import hippodamia.schema.mongodb as mongodb
import hippodamia.schema.webserver as webserver
import hippodamia.schema.email as email


def get_schema():
    schema = {
        "monitoringservice": {
            "description": "Hippodamia observes the state of all registered microservices (aka watch dog).",
            "type": "object",
            "properties": {
                "topics": topics.get_schema(),
                "service-timings": service_timings.get_schema(),
                "agent-timings": agent_timings.get_schema(),
                "expected-services": expected_services.get_schema(),
                "mongodb": mongodb.get_schema(),
                "webserver": webserver.get_schema(),
                "email": email.get_schema()
            },
            "required": ["topics", "agent-timings", "service-timings", "expected-services"],
            "additionalProperties": False
        }
    }
    return schema




def get_schema():
    schema = {
                "type": "object",
                "description": "intervals in seconds for different tasks of the monitoring agent - transmitted in the "
                               "onboarding response message",
                "properties": {
                    "send-ping": {
                        "description": "interval in seconds for the scheduler to send a ping message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 60 seconds.",
                        "type": "integer"
                    },
                    "send-runtime": {
                        "description": "interval in seconds for the scheduler to send a runtime message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 500 seconds.",
                        "type": "integer"
                    },
                    "send-config": {
                        "description": "interval in seconds for the scheduler to send a config message to the "
                                       "monitoring service. 0 deactivates the scheduler. example: 3600 seconds.",
                        "type": "integer"
                    },
                    "expect-heartbeat": {
                        "description": "maximum interval in seconds until a heartbeat message from the monitoring"
                                       "service should be received. If violated, the agent start the re-onboarding"
                                       "procedures. 0 deactivates this feature. example: 500 seconds.",
                        "type": "integer"
                    },
                },
                "required": ["send-ping", "send-runtime", "send-config", "expect-heartbeat"]
            }
    return schema



import smtplib
from email.mime.text import MIMEText
from pelops.logging.mylogger import get_child


class SendEmail:
    """
    Classic publish to email implementation.

    config yaml entries:

    email:
            address: smtp.zoho.com  # url of smtp server
            port: 465  # port of smtp server
            from: name123@sender123.com  # sender email address
            credentials-file: ~/credentials.yaml  # a yaml config file with one root entry as described below

    credentials yaml file:

    monitoringservice:
        email:
            config:
                username: username  # username for smtp server
                password: password  # password for smtp server
    """
    _counter = None
    _config = None
    _address = None  # url of smtp server
    _port = None  # port of smtp server
    _username = None  # username for smtp server
    _password = None  # password for smtp server
    _from = None  # sender email address

    def __init__(self, config, logger):
        """
        Constructor

        :param config: yaml config structure
        :param logger: logger instance from the parent. a child will be spawned
        """
        self._config = config
        self._logger = get_child(logger, self.__class__.__name__)
        self._logger.info("__init__")
        self._logger.debug("__init__: {}".format(self._config))
        self._address = self._config["address"]
        self._port = self._config["port"]
        self._username = self._config["username"]
        self._password = self._config["password"]
        self._from = self._config["from"]

        if not self._test_connection():
            self._logger.error("connection test failed")
            raise RuntimeError("connection test failed")
        else:
            self._logger.debug("connection test success")
        self._counter = 0

    def _test_connection(self):
        try:
            server = smtplib.SMTP_SSL(host=self._address, port=self._port)
            server.set_debuglevel(0)
            server.login(self._username, self._password)
            server.noop()
            server.quit()
            return True
        except smtplib.SMTPException as exception:
            self._logger.error("test_connection - failed: {}.".format(exception))
        return False

    def send_message(self, subject, message, to):
        self._logger.info("send_message - to: {}, subject: {}, lem(message): {}".format(to, subject, len(message)))
        msg = MIMEText(message, 'html')
        if isinstance(to, str):
            _to = [to]
        else:
            _to = to

        msg['To'] = ', '.join(_to)
        msg['From'] = self._from
        msg['Subject'] = subject

        try:
            server = smtplib.SMTP_SSL(host=self._address, port=self._port)
            server.set_debuglevel(0)
            server.login(self._username, self._password)
            server.send_message(msg)
        except smtplib.SMTPException as exception:
            self._logger.error("send_message - failed: {}.".format(exception))
            raise

        server.quit()
        self._counter += 1

    def get_stats(self):
        stats = {
            "count": self._counter,
            "address": self._address,
            "port": self._port,
            "from": self._from
        }
        return stats



class MyMongoClient:
    _config = None
    _logger = None

    def __init__(self, config, logger):
        self._config = config
        self._logger = logger

    def start(self):
        pass

    def stop(self):
        pass

    def write_shadow(self, document):
        pass



def get_schema():
    schema = {
        "type": "object",
        "description": "persistance database",
        "properties": {
            "credentials-file": {
                "type": "string",
                "example": "~/credentials.yaml",
                "description": "credentials for mongodb"
            },
            "mongodb-address": {
                "type": "string",
                "example": "localhost",
                "description": "host name"
            },
            "mongodb-port": {
                "type": "integer",
                "example": 21017,
                "description": "port"
            },
            "database": {
                "type": "string",
                "example": "hippodamia",
                "description": "database name"
            },
            "mongodb-user": {
                "description": "User name for mongodb",
                "type": "string"
            },
            "mongodb-password": {
                "description": "Password for mongodb",
                "type": "string"
            },
            "log-level": {
                "type": "string",
                "example": "DEBUG",
                "description": "log level for mongodb client related messages",
                "enum": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
            }
        },
        "required": ["mongodb-address", "mongodb-port", "database", "log-level"],
        "additionalProperties": False
    }
    return schema


from hippodamia.hierarchivalview.treeentry import Entry
from hippodamia.enums import ViewDetails
from hippodamia.enums import Health
from hippodamia.enums import Necessity
from collections import deque


class Leaf(Entry):
    gid = None
    shadow = None
    breadcrumbs = None

    def __init__(self, shadow):
        Entry.__init__(self, shadow.properties.name)
        self.shadow = shadow
        self.gid = self.shadow.properties.gid
        self.breadcrumbs = deque()
        self.update()

    def update(self):
        self.health = self.shadow.properties.health
        self.name = self.shadow.properties.name
        self.breadcrumbs = deque([self.shadow.properties.location, self.shadow.properties.room,
                                  self.shadow.properties.device])
        result = self.health
        if self.shadow.properties.necessity != Necessity.REQUIRED:
            result = min(result, Health.YELLOW)
        return result

    def get_state(self):
        return self.shadow.get_state_id().name

    def get_health(self):
        return self.health.name

    def pformat(self, details=ViewDetails.NONE, filters=None):
        result = [self._name2str()]

        if details & ViewDetails.HEALTH == ViewDetails.HEALTH:
            result.append(self.get_health())
        if details & ViewDetails.STATE == ViewDetails.STATE:
            result.append(self.get_state())
        if details & ViewDetails.GID == ViewDetails.GID:
            result.append("gid:")
            result.append(self.gid)

        result = " ".join(result)
        return result

    def to_dict(self):
        result = {
            "type": self.__class__.__name__,
            "name": self._name2str(),
            "health": self.get_health(),
            "state": self.get_state(),
            "gid": self.gid
        }
        return result



import threading
import json

from pelops.abstractmicroservice import AbstractMicroservice
import hippodamia
from hippodamia.agentshadow.agentshadowfactory import AgentShadowFactory
from hippodamia.mymongoclient import MyMongoClient
from hippodamia.heartbeat import Heartbeat
from hippodamia.commands import CommandFactory
from hippodamia.schema.monitoringservice import get_schema
from hippodamia.hierarchivalview.hierarchicalview import HierarchicalView
from hippodamia.enums import Enforcement
from hippodamia.onboarding import Onboarding
from hippodamia.flask.server import FlaskServer
from hippodamia.agentshadow.states.state_ids import state_ids
from hippodamia.email.emailnotification import EmailNotification


class Monitoringservice(AbstractMicroservice):
    _version = hippodamia.version

    _protocol_version = 1

    _agentshadows = None
    _agentshadow_archive = None
    _asfactory = None
    _mongo_client = None
    _heartbeat = None
    _hierarchical_view = None
    _onboarding = None
    _flask_server = None
    _email_notification = None

    _enforcement = None

    _topic_onboarding_request = None
    _topic_cmd_ping = None
    _topic_cmd_runtime = None
    _topic_cmd_config = None
    _topic_cmd_reonboarding = None
    _topic_cmd_end = None

    _initial_reonboarding_request = None

    def __init__(self, config, mqtt_client=None, logger=None, stdout_log_level=None, no_gui=None):
        """
        Constructor - creates the services and the tasks

        :param config: config yaml structure
        :param mqtt_client: mqtt client instance
        :param logger: logger instance
        :param no_gui: if False create and control a ui instance
        :param stdout_log_level: if set, a logging handler with target sys.stdout will be added
        """
        forced_monitoring_agent_publishloglevel = False
        try:
            ll = config["monitoring-agent"]["publish-log-level"]
            if ll.upper() == "DEBUG" or ll.upper() == "INFO":
                config["monitoring-agent"]["publish-log-level"] = "WARNING"
                forced_monitoring_agent_publishloglevel = True
        except KeyError:
            pass

        AbstractMicroservice.__init__(self, config, "monitoringservice", mqtt_client, logger,
                                      logger_name=__class__.__name__, stdout_log_level=stdout_log_level, no_gui=no_gui)

        if forced_monitoring_agent_publishloglevel:
            self._logger.warning("forced 'monitoring-agent.publish-log-level' of monitoring agent to be 'warning' or "
                                 "higher. as every action of a monitoring agent results in corresponding actions in "
                                 "the monitoring service which in this case would trigger logging events in its "
                                 "monitoring agent which would then trigger again an update in the monitoring "
                                 "service ...")
        try:
            self._mongo_client = MyMongoClient(self._config["mongodb"], self._logger)
        except KeyError:
            self._logger.info("no entry 'mongodb' in config -> no mongodb will be used")

        self._heartbeat = Heartbeat(self._config["topics"]["heartbeat"],
                                    self._config["service-timings"]["heartbeat_interval"],
                                    self._mqtt_client, self._logger)

        self._topic_onboarding_request = self._config["topics"]["onboarding-request"]
        self._topic_cmd_ping = self._config["topics"]["command-ping-on-request"]
        self._topic_cmd_runtime = self._config["topics"]["command-runtime-on-request"]
        self._topic_cmd_config = self._config["topics"]["command-config-on-request"]
        self._topic_cmd_reonboarding = self._config["topics"]["command-reonboarding"]
        self._topic_cmd_end = self._config["topics"]["command-end"]

        self._asfactory = AgentShadowFactory(self._config["topics"], self._config["service-timings"],
                                             self._config["agent-timings"],
                                             self._config["expected-services"]["services"],
                                             self._mongo_client, self._mqtt_client, self._logger)
        self._agentshadows = {}
        self._asfactory.add_predefined_shadows(self._agentshadows)
        self._agentshadow_archive = []

        self._hierarchical_view = HierarchicalView(self._agentshadows, self._logger)

        self._enforcement = Enforcement.get_enum(self._config["expected-services"]["enforcement"])
        self._onboarding = Onboarding(self._protocol_version, self._topic_onboarding_request, self._enforcement,
                                      self._agentshadows, self._asfactory, self._mqtt_client, self._logger)

        try:
            self._email_notification = EmailNotification(self._config["email"], self._logger, self._mqtt_client,
                                                         self._agentshadows, self._heartbeat, self._hierarchical_view,
                                                         self._onboarding, self._agentshadow_archive, self._asfactory)
        except KeyError:
            self._logger.info("no entry 'email' in config -> email notification is disabled")

        if "webserver" in self._config:
            self._flask_server = FlaskServer(self._config["webserver"], self._logger, self._agentshadows,
                                             self._agentshadow_archive, self._heartbeat, self._hierarchical_view,
                                             self._onboarding, self.send_cmd_ping, self.send_cmd_runtime,
                                             self.send_cmd_config, self.send_cmd_reonboarding, self.send_cmd_end,
                                             self._full_config, self._mqtt_client, self.archive_agentshadows,
                                             self.purge_agentshadow_archive, self.acknowledge_log,
                                             self._email_notification)
        else:
            self._logger.info("no entry 'webserver' in config -> no webserver will be created")

        factory = CommandFactory(self._version, self._config, self._agentshadows, self._heartbeat,
                                 self._hierarchical_view, self._onboarding, self.send_cmd_ping, self.send_cmd_runtime,
                                 self.send_cmd_config, self.send_cmd_reonboarding, self.send_cmd_end,
                                 self._agentshadow_archive, self._email_notification, self._logger)
        factory.add_commands(self._add_ui_command)

    def _start(self):
        if self._flask_server is not None:
            self._flask_server.start()
        if self._mongo_client is not None:
            self._mongo_client.start()
        self._heartbeat.start()
        for k, v in self._agentshadows.items():
            v.start()
        self._hierarchical_view.start()
        if self._email_notification is not None:
            self._email_notification.start()
        self._onboarding.start()
        self.send_cmd_reonboarding()

    def _stop(self):
        self._heartbeat.stop()
        self._onboarding.stop()
        if self._email_notification is not None:
            self._email_notification.stop()
        self.send_cmd_end(gid=None)  # inform everyone that the service is going down
        self._hierarchical_view.stop()
        for k, v in self._agentshadows.items():
            print("stopping {}".format(k))
            v.stop()
        if self._mongo_client is not None:
            self._mongo_client.stop()
        if self._flask_server is not None:
            self._flask_server.stop()

    def archive_agentshadows(self):
        """remove all agentshadows in state 'ARCHIVECANDIDATE' from agentshadows, deactivate them and move them to
        a seperate list"""

        self._logger.info("archivation of shadows - start")
        self._logger.debug("collect list of to be archived shadows")
        candidates = {}
        for gid, agentshadow in self._agentshadows.items():
            if agentshadow.get_state_id() == state_ids.ARCHIVECANDIDATE:
                candidates[gid] = agentshadow

        self._logger.debug("archive these shadows: {}".format(candidates.keys()))

        for gid, agentshadow in candidates.items():
            agentshadow.prepare_for_archivation()
            del self._agentshadows[gid]
            self._agentshadow_archive.append(agentshadow)

        self._logger.info("archivation of shadows (added: {}) - finished".format(len(candidates)))
        self._asfactory.add_predefined_shadows(self._agentshadows)

        return len(candidates)

    def purge_agentshadow_archive(self):
        counter = len(self._agentshadow_archive)
        self._logger.info("purging of agendshadow archive - removing {} entries".format(counter))
        self._agentshadow_archive.clear()
        return counter

    @classmethod
    def _get_description(cls):
        return "Hippodamia observes the state of all registered microservices (aka watch dog)."

    @classmethod
    def _get_schema(cls):
        return get_schema()

    def runtime_information(self):
        info = {}
        for gid, shadow in self._agentshadows.items():
            entry = {
                "gid": gid,
                "state": shadow.get_state_id().name,
                "health": shadow.properties.health.name,
                "necessity": shadow.properties.necessity.name,
                "properties": shadow.properties.name
            }
            info[gid] = entry
        return info

    def config_information(self):
        return {}

    def acknowledge_log(self, gid=None):
        counter = 0
        if gid is None:
            self._logger.info("acknowledge_log: all agentshadows")
            for g, a in self._agentshadows.items():
                self._logger.debug("acknowledge_log: gid: {}".format(g))
                a.log_archive.reset_max_level()
                counter += 1
        elif not gid in self._agentshadows:
            self._logger.info("acknowledge_log: unknown gid")
        else:
            self._logger.info("acknowledge_log: gid: {}".format(gid))
            self._agentshadows[gid].log_archive.reset_max_level()
            counter += 1
        return counter

    @staticmethod
    def _render_message(request, gid):
        """
        {
            "request": "request",
            "gid": 1
        }
        """
        message = {
            "request": request
        }
        if gid is not None:
            message["gid"] = gid
        return json.dumps(message)

    def send_cmd_ping(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_ping: send request to all monitored services")
        else:
            self._logger.info("send_cmd_ping: send request to gid: {}".format(gid))
        message = self._render_message("ping", gid)
        self._mqtt_client.publish(self._topic_cmd_ping, message)

    def send_cmd_runtime(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_runtime: send request to all monitored services")
        else:
            self._logger.info("send_cmd_runtime: send request to gid: {}".format(gid))
        message = self._render_message("runtime", gid)
        self._mqtt_client.publish(self._topic_cmd_runtime, message)

    def send_cmd_config(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_config: send request to all monitored services")
        else:
            self._logger.info("send_cmd_config: send request to gid: {}".format(gid))
        message = self._render_message("config", gid)
        self._mqtt_client.publish(self._topic_cmd_config, message)

    def send_cmd_reonboarding(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_reonboarding: send request to all monitored services")
        else:
            self._logger.info("send_cmd_reonboarding: send request to gid: {}".format(gid))
        message = self._render_message("reonboarding", gid)
        self._mqtt_client.publish(self._topic_cmd_reonboarding, message)

    def send_cmd_end(self, gid=None):
        if gid is None:
            self._logger.info("send_cmd_end: send request to all monitored services")
        else:
            self._logger.info("send_cmd_end: send request to gid: {}".format(gid))
        message = self._render_message("end", gid)
        self._mqtt_client.publish(self._topic_cmd_end, message)


def standalone():
    Monitoringservice.standalone()


if __name__ == "__main__":
    Monitoringservice.standalone()



from hippodamia.hierarchivalview.treebranch import Branch


class Root(Branch):
    def __init__(self, leafs):
        Branch.__init__(self, name="__root__")
        self._build_tree(leafs)

    def _build_tree(self, leafs):
        new_children = {}

        for gid, leaf in leafs.items():
            breadcrumbs = leaf.breadcrumbs.copy()
            self.add_leaf(leaf, breadcrumbs, new_children)

        self.children = new_children
        self.update()

    def to_dict(self):
        result = Branch.to_dict(self)
        result["type"] = Branch.__name__
        return result



import datetime
import collections
import pprint


class _ArchiveEntry:
    messages = None
    counter = None
    time = None
    _maxlen = None

    def __init__(self, maxlen=10):
        self._maxlen = maxlen
        self.messages = collections.deque(maxlen=self._maxlen)
        self.counter = 0
        self.time = None

    def add(self, message):
        self.messages.append(message)
        self.counter += 1
        self.time = datetime.datetime.now()

    def get_last_message(self):
        try:
            return self.messages[-1]
        except IndexError:
            return None

    def get_stats(self, full):
        try:
            str_time = self.time.strftime("%Y-%m-%dT%H:%M:%S")
        except AttributeError:
            str_time = None

        result = {
            "counter": self.counter,
            "stored": len(self.messages),
            "time": str_time,
            "maxlen": self._maxlen,
            "last_message": self.get_last_message(),
        }
        if full:
            result["messages"] = self.messages
        return result


class MessageArchive:
    logger = None
    request = None
    config = None
    runtime = None
    ping = None
    end = None

    def __init__(self):
        self.logger = _ArchiveEntry(maxlen=1)
        self.request = _ArchiveEntry(maxlen=5)
        self.config = _ArchiveEntry(maxlen=5)
        self.runtime = _ArchiveEntry(maxlen=10)
        self.ping = _ArchiveEntry(maxlen=50)
        self.end = _ArchiveEntry(maxlen=5)

    def archive_ping(self, message):
        self.ping.add(message)

    def archive_config(self, message):
        self.config.add(message)

    def archive_runtime(self, message):
        message["process_time_percent"] = message["process-time"] / message["service-uptime"]
        self.runtime.add(message)

    def archive_end(self, message):
        self.end.add(message)

    def archive_logger(self, message):
        self.logger.add(message)

    def archive_request(self, message):
        self.request.add(message)

    def get_stats(self, full=True):
        result = {
            "request": self.request.get_stats(full),
            "ping": self.ping.get_stats(full),
            "runtime": self.runtime.get_stats(full),
            "config": self.config.get_stats(full),
            "logger": self.logger.get_stats(full),
            "end": self.end.get_stats(full)
        }
        return result

    def get_string_stats(self, key):
        def _generate_output(archive):
            output = "stats:\n"
            output += " - time: {}\n".format(archive["time"])
            output += " - counter: {}\n".format(archive["counter"])
            output += " - stored: {}\n".format(archive["stored"])
            output += " - maxlen: {}\n".format(archive["maxlen"])
            output += "messages:\n"
            for m in archive["messages"]:
                output += "{}\n".format(pprint.pformat(m, indent=4))
            output += "\n"
            return output

        archive = self.get_stats(full=True)
        if key in archive:
            output = _generate_output(archive[key])
        elif key == "all":
            output = ""
            keys = list(archive.keys())
            keys.sort()
            for k in keys:
                output += "'{}':\n".format(k)
                output += _generate_output(archive[k])
                output += "\n"
        else:
            output = "unknown key '{}'\n".format(key)

        return output



from hippodamia.agentshadow.agentshadow import AgentShadow


class AgentShadowFactory:
    _topics = None
    _service_timings = None
    _agent_timings = None
    _mongo_client = None
    _mqtt_client = None
    _logger = None
    _preload_config_list = None
    _observer_state_events = None
    _observer_log_events = None

    def __init__(self, topics, service_timings, agent_timings, preload_config_list, mongo_client,
                 mqtt_client, logger):
        self._topics = topics
        self._service_timings = service_timings
        self._agent_timings = agent_timings
        self._mongo_client = mongo_client
        self._mqtt_client = mqtt_client
        self._logger = logger
        self._preload_config_list = preload_config_list

    def add_predefined_shadows(self, agentshadows):
        def _already_loaded(preload_config):
            for entry in agentshadows.values():
                if (entry.properties.location == preload_config["location"] and
                        entry.properties.room == preload_config["room"] and
                        entry.properties.device == preload_config["device"] and
                        entry.properties.name == preload_config["name"]):
                    return True
            return False

        def _add_entry(preload_config):
            gid = str(preload_config["gid"])
            shadow = self.new_agentshadow(preload_config=preload_config)
            shadow.register_state_transistion_event_observer(self._observer_state_events)
            shadow.register_log_event_observer(self._observer_log_events)
            agentshadows[gid] = shadow

        self._logger.info("adding predefined shadows - start")
        for preload_config in self._preload_config_list:
            if _already_loaded(preload_config):
                self._logger.debug("agentshadow already loaded for {}".format(preload_config))
                continue
            else:
                self._logger.debug("adding preloadconfig for agentshadow {}".format(preload_config))
                _add_entry(preload_config)
        self._logger.info("adding predefined shadows - finished")

    def new_agentshadow(self, preload_config=None, gid=None):
        if preload_config is None and gid is None:
            raise ValueError("AgentShadowFactory.create_agentshadow - preload_config and gid are None")
        shadow = AgentShadow(self._topics, self._service_timings, self._agent_timings,
                             self._mongo_client, self._mqtt_client, self._logger, preload_config=preload_config,
                             gid=gid)
        shadow.register_state_transistion_event_observer(self._observer_state_events)
        shadow.register_log_event_observer(self._observer_log_events)
        return shadow

    @staticmethod
    def get_string_list(agentshadows):
        m = ""
        gids = list(agentshadows.keys())
        gids.sort()
        for gid in gids:
            shadow = agentshadows[gid]
            m += "gid: {}, state: {}, health:{}, necessity: {}, name: {}, type:{}, log-entries: {}\n"\
                .format(gid,
                        shadow.get_state_id().name,
                        shadow.properties.health.name,
                        shadow.properties.necessity.name,
                        shadow.properties.name,
                        shadow.properties.type,
                        shadow.log_archive.flat_list.counter)
        return m

    def register_state_transistion_event_observer(self, observer, agentshadows):
        self._observer_state_events = observer
        for shadow in agentshadows.values():
            shadow.register_state_transistion_event_observer(self._observer_state_events)

    def register_log_event_observer(self, observer, agentshadows):
        self._observer_log_events = observer
        for shadow in agentshadows.values():
            shadow.register_log_event_observer(self._observer_log_events)




from enum import Enum


class state_ids(Enum):
    ONBOARDING = 1
    ONBOARDED = 2
    ACTIVE = 3
    INACTIVE = 4 
    STOPPED = 5
    ARCHIVECANDIDATE = 6
    ARCHIVED = 7
    PREDEFINED = 8
    LOST = 9





def get_schema():
    schema = {
        "type": "object",
        "description": "pre-configure which services are expected to be onboarded",
        "properties": {
            "enforcement": {
                "type": "string",
                "example": "none",
                "description": "spontaneously appearing services are either [none]-accepted, [ignore]-ignored, "
                               "[strict]-leads to a error log entry",
                "enum": ["none", "ignore", "strict"]
            },
            "services": {
                "type": "array",
                "description": "list of expected services",
                "items": {
                    "type": "object",
                    "properties": {
                        "gid": {
                            "type": ["string", "integer"],
                            "example": "1",
                            "description": "gid of the service"
                        },
                        "name": {
                            "type": "string",
                            "example": "Thermostat GUI",
                            "description": "name of the service (optional)"
                        },
                        "location": {
                            "type": "string",
                            "example": "vienna",
                            "description": "location of the service (optional)"
                        },
                        "room": {
                            "type": "string",
                            "example": "living room",
                            "description": "room of the service (optional)"
                        },
                        "device": {
                            "type": "string",
                            "example": "thermostat",
                            "description": "device the service is running on (optional)"
                        },
                        "necessity": {
                            "type": "string",
                            "example": "required",
                            "description": "configure if the expected service is [required] or [optional]",
                            "enum": ["required", "optional"]
                        }
                    },
                    "required": ["gid", "necessity"],
                    "additionalProperties": False
                }
            }
        },
        "required": ["enforcement", "services"],
        "additionalProperties": False
    }
    return schema


import pprint
from hippodamia.email.sendemail import SendEmail
from hippodamia.email.digest import Digest
from hippodamia.email.spontaneous import Spontaneous
from pelops.logging.mylogger import get_child


class EmailNotification:
    _config = None
    _logger = None
    _send_email = None

    _spontaneous = None
    _digest = None

    def __init__(self, config, logger, mqtt_client, agentshadows, heartbeat, hierarchical_view, onboarding, archive,
                 agentfactory):
        self._config = config
        self._logger = get_child(logger, self.__class__.__name__)
        self._logger.info("__init__ - begin")

        self._send_email = SendEmail(self._config["config"], self._logger)

        try:
            self._digest = Digest(self._config["digest"], self._send_email.send_message, self._logger, mqtt_client,
                                  agentshadows, heartbeat, hierarchical_view, onboarding, archive)
        except KeyError:
            self._logger.info("Digest is deactivated")

        try:
            self._spontaneous = Spontaneous(self._config["spontaneous"], self._send_email.send_message, self._logger,
                                            agentshadows, agentfactory)
        except KeyError:
            self._logger.info("Spontaneous is deactivated")

        if self._digest is None and self._spontaneous is None:
            self._logger.error("Digest and Spontaneous are deactived - at least one must be set.")
            raise RuntimeError("Digest and Spontaneous are deactived - at least one must be set.")

        self._logger.info("__init__ - completed")

    def start(self):
        self._logger.info("start - begin")
        if self._digest is not None:
            self._digest.start()
        if self._spontaneous is not None:
            self._spontaneous.start()
        self._logger.info("start - completed")

    def stop(self):
        self._logger.info("stop - begin")
        if self._digest is not None:
            self._digest.stop()
        if self._spontaneous is not None:
            self._spontaneous.stop()
        self._logger.info("stop - completed")

    def get_stats(self):
        stats = {
            "sendemail": self._send_email.get_stats()
        }
        if self._digest is not None:
            stats["digest"] = self._digest.get_stats()
        if self._spontaneous is not None:
            stats["spontaneous"] = self._spontaneous.get_stats()
        return stats

    def get_string_stats(self):
        text = "stats:\n"
        text += pprint.pformat(self.get_stats(), indent=4)
        return text



from hippodamia.agentshadow.states.state_ids import state_ids


def get_schema():
    schema = {
            "description": "Email notification service for hippodamia. At least 'digest' or 'spontaneous' must be "
                           "present to be operational.",
            "type": "object",
            "properties": {
                "config": _config_schema(),
                "digest": _digest_schema(),
                "spontaneous": _spontaneous_schema(),
            },
            "required": ["config"],
            "additionalProperties": False
        }
    return schema


def _config_schema():
    schema = {
        "description": "Smtp server configuration",
        "type": "object",
        "properties": {
            "address": {
                "description": "url",
                "type": "string"
            },
            "port": {
                "description": "port (e.g.: 587)",
                "type": "integer",
                "minimum": 0,
                "maximum": 65535
            },
            "credentials-file": {
                "description": "File containing the credentials (optional).",
                "type": "string"
            },
            "username": {
                "description": "username for smtp server",
                "type": "string"
            },
            "password": {
                "description": "password for smtp server",
                "type": "string"
            },
            "from": {
                "description": "sender e-mail address",
                "type": "string"
            }
        },
        "required": ["address", "port", "username", "password", "from"],
        "additionalProperties": False
    }
    return schema


def _digest_schema():
    schema = {
        "description": "Regularly send a digest notification with the most important data. At least 'at' or 'start-day'"
                       " must be given to be operational.",
        "type": "object",
        "properties": {
            "subject": {
                "description": "subject of the email notification",
                "type": "string"
            },
            "pre-text": {
                "description": "text to be added before the auto-generated content",
                "type": "string"
            },
            "post-text": {
                "description": "text to be added after the auto-generated content",
                "type": "string"
            },
            "to": {
                "description": "receiver e-mail address",
                "oneOf": [
                    {"type": "string"},
                    {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }
                ]
            },
            "repeat": {
                "description": "'...' repeat send digest",
                "type": "string",
                "enum": ["HOURLY", "DAILY", "WEEKLY"]
            },
            "at": {
                "description": "send digest notification at this time. in case of repeat=='HOURLY', delay until this "
                               "time. if omited, use time at startup.",
                "type": "string",
                "pattern": "^(0[0-9]|1[0-9]|2[0-3]):[0-5][0-9]$",
                "example": "12:34"
            },
            "start-day": {
                "description": "similar to 'at'. send digest notification on this day. in case of repeat=='HOURLY' or"
                               " repeat=='DAILY', delay until this day. if omitted, use day at startup.",
                "type": "string",
                "enum": ["MONDAY", "TUESDAY", "WEDNESDAY", "THURSDAY", "FRIDAY", "SATURDAY", "SUNDAY"]
            },
            "send-at-startup": {
                "description": "send a digest immediately after startup - optional (default: False)",
                "type": "boolean"
            }
        },
        "required": ["subject", "to", "repeat"],
        "additionalProperties": False
    }
    return schema


def _spontaneous_schema():
    state_names = [e.name for e in state_ids]

    schema = {
        "description": "Send a notification upon an event. Events can either be a state change or reception of a "
                       "log entry with a given log-level.",
        "type": "object",
        "properties": {
            "subject": {
                "description": "subject of the email notification",
                "type": "string"
            },
            "pre-text": {
                "description": "text to be added before the auto-generated content",
                "type": "string"
            },
            "post-text": {
                "description": "text to be added after the auto-generated content",
                "type": "string"
            },
            "to": {
                "description": "receiver e-mail address",
                "oneOf": [
                    {"type": "string"},
                    {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }
                ]
            },
            "delay": {
                "description": "Delay generation of notification by n-seconds. This avoids generation of alarm showers "
                               "for events that have the same root-cause.",
                "type": "number",
                "minimum": 0
            },
            "state_events": {
                "description": "Set of state transition events.",
                "type": "array",
                "items": {
                    "description": "The transition from state 'A' to state 'B' will trigger a notification event.",
                    "type": "object",
                    "properties": {
                        "from": {
                            "description": "start state",
                            "type": "string",
                            "enum": state_names
                        },
                        "to": {
                            "description": "target state",
                            "type": "string",
                            "enum": state_names
                        }
                    },
                    "required": ["from", "to"],
                    "additionalProperties": False
                }

            },
            "log_events": {
                "description": "A log message with the given log-level will trigger a notification event.",
                "type": "array",
                "items": {
                    "type": "string",
                    "enum": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
                }
            }
        },
        "required": ["subject", "to", "delay"],
        "additionalProperties": False
    }
    return schema




import datetime
from tantamount.astate import AState
import pelops.logging.mylogger


class AMonitoringState(AState):
    logger = None
    counter_on_exit = 0
    counter_on_entry = 0
    timestamp_on_entry = None
    timestamp_on_exit = None
    set_health = None
    health = None
    reset_max_level = None

    def __init__(self, id, set_health, health, reset_max_level, logger, logger_name, group_id):
        AState.__init__(self, id, groupid=group_id)
        self.logger = pelops.logging.mylogger.get_child(logger, logger_name)
        self.logger.debug("__init__ ")
        self.set_health = set_health
        self.health = health
        self.reset_max_level = reset_max_level

    def on_entry(self):
        self.logger.info("on_entry")
        self.counter_on_entry += 1
        self.timestamp_on_exit = None
        self.timestamp_on_entry = datetime.datetime.now()

        self.set_health(self.health)
        next = self._on_entry()

        if next is not None:
            self.logger.debug("on_entry - event: {}".format(next.name))
        return next

    def _on_entry(self):
        return None

    def on_exit(self):
        self.logger.info("on_exit")
        self.counter_on_exit += 1
        self.timestamp_on_exit = datetime.datetime.now()

        self._on_exit()

    def _on_exit(self):
        pass

    def get_stats(self):
        try:
            duration = self.timestamp_on_exit - self.timestamp_on_entry
        except TypeError:
            try:
                duration = datetime.datetime.now() - self.timestamp_on_entry
            except TypeError:
                duration = "-"

        try:
            str_time_on_entry = self.timestamp_on_entry.strftime("%Y-%m-%dT%H:%M:%S")
        except AttributeError:
            str_time_on_entry = None

        try:
            str_time_on_exit = self.timestamp_on_entry.strftime("%Y-%m-%dT%H:%M:%S")
        except AttributeError:
            str_time_on_exit = None

        try:
            float_duration = duration.total_seconds()
        except AttributeError:
            float_duration = None

        stats = {
            "counter_on_entry": self.counter_on_entry,
            "timestamp_on_entry": str_time_on_entry,
            "counter_on_exit": self.counter_on_exit,
            "timestamp_on_exit": str_time_on_exit,
            "duration": float_duration
        }
        return stats



from werkzeug.serving import make_server
import threading


class ServerThread(threading.Thread):
    """
    https://stackoverflow.com/questions/15562446/how-to-stop-flask-application-without-using-ctrl-c/45017691#45017691
    """

    _server = None
    _context = None

    def __init__(self, app, port, host="0.0.0.0"):
        threading.Thread.__init__(self, name="{} ({}:{})".format(__class__.__name__, host, port))
        self._server = make_server(app=app, port=port, host=host)
        self._context = app.app_context()
        self._context.push()

    def run(self):
        self._server.serve_forever()

    def stop(self):
        self._server.shutdown()
